#ifndef USE_KDP_CAMERA
#ifdef V2K_ENABLE
/*
 * @name : sys_camera.c
 * @brief : camera image capture (abstract) bus driver
 *
 * Copyright (C) 2019 Kneron, Inc. All rights reserved.
 *
 */
#include "config/board_kdp520.h"
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "base.h"
#include "utility.h"
#include "kdrv_scu_ext.h"
#include "framework/init.h"
#include "framework/kdp_list.h"
#include "framework/utils.h"
#include "framework/bitops.h"
#include "framework/framework_driver.h"
#include "framework/event.h"
#include "framework/framework_errno.h"
#include "framework/v2k_image.h"
#include "media/camera/sensor.h"
#include "media/camera/sys_camera.h"
#include "media/v2k_subdev.h"
#include "media/v2k_ioctl.h"
#include "media/videobuf_core.h"
#include "kmdw_console.h"


#define CAMERA_NUMS_LIMIT 1//2
// #define MAP_MAX_NUM 2
// static unsigned long device_map[BITS_TO_LONGS(MAP_MAX_NUM)];

#define SYS_CAMERA_BUG_0720

struct kdp_list hosts = { &(hosts), &(hosts) };
struct kdp_list devices = { &(devices), &(devices) };

extern struct kdp_list g_klist_drivers;

static int _sys_camera_probe(struct sys_camera_host *, struct sys_camera_device *);
static int _sys_camera_probe_finish(struct sys_camera_device *);
//static void _scan_add_host(struct sys_camera_host *);

static inline struct sys_camera_desc *to_sys_camera_desc(
        const struct sys_camera_device *cam_d) {
    return cam_d->sdesc;
}

static inline struct pin_context *to_sys_camera_control(
        const struct sys_camera_device *cam_d) {
    return cam_d->child_pin;
}
        
static const struct sys_camera_pixelfmt sys_camera_fmt[] =
{
    {
        .fourcc             = V2K_PIX_FMT_RGB565,
        .bits_per_sample    = 8,
        .packing            = SYS_CAMERA_PACKING_2X8_PADHI,
        //.order              = SYS_CAMERA_ORDER_LE,
        .layout             = SYS_CAMERA_LAYOUT_PACKED,
    },
    {
        .fourcc             = V2K_PIX_FMT_RAW10,
        .bits_per_sample    = 16,
        .packing            = SYS_CAMERA_PACKING_NONE,
        //.order              = SYS_CAMERA_ORDER_LE,
        .layout             = SYS_CAMERA_LAYOUT_PACKED,
    },    
    {
        .fourcc             = V2K_PIX_FMT_RAW8,
        .bits_per_sample    = 8,
        .packing            = SYS_CAMERA_PACKING_NONE,
        //.order              = SYS_CAMERA_ORDER_LE,
        .layout             = SYS_CAMERA_LAYOUT_PACKED,
    },
};

const struct sys_camera_pixelfmt *sys_camera_get_fmtdesc(unsigned int fourcc) {
    int n = ARRAY_SIZE(sys_camera_fmt);
    int i;
    for (i = 0; i < n; ++i)
        if (sys_camera_fmt[i].fourcc == fourcc)
            return &sys_camera_fmt[i];
    return NULL;
}

int sys_camera_bytes_per_line(unsigned int fourcc, unsigned int width) {
    const struct sys_camera_pixelfmt *p = sys_camera_get_fmtdesc(fourcc);
    switch (p->packing) {
        case SYS_CAMERA_PACKING_NONE:
            return width * p->bits_per_sample / 8;
        case SYS_CAMERA_PACKING_2X8_PADHI:
        case SYS_CAMERA_PACKING_2X8_PADLO:
            return width * 2;
        default:;
    }
    return 0;
}

struct sys_camera_host *to_sys_camera_host(const struct pin_context *pin_ctx) {
    struct v2k_pin_obj *pin_obj = framework_drv_get_drvdata(pin_ctx);
    return container_of(pin_obj, struct sys_camera_host, pin_obj);
    //return container_of(pin_ctx, struct sys_camera_host, pin_ctx);
}

struct v2k_subdev *sys_camera_to_subdev(const struct sys_camera_device *cam_d) {
    struct pin_context *pin_ctx = to_sys_camera_control(cam_d);
    return framework_drv_get_drvdata(pin_ctx);
}

int sys_camera_host_register(struct sys_camera_host *cam_h)
{
    int ret;
    struct sys_camera_host *sch;
    
    DSG("[%s]", __func__);
    
    if (!cam_h || !cam_h->ops ||
        !cam_h->ops->set_fmt ||
        !cam_h->ops->querycap ||
        !cam_h->ops->init_videobuf)
        return -KDP_FRAMEWORK_ERRNO_INVALA;

    struct kdp_list *head = &hosts;

    for (sch = kdp_list_first_entry(head, typeof(*sch), list);
         &sch->list != head;
         sch = kdp_list_next_entry(sch, list))
    {
        DSG("[%s] sch->host_id=%d cam_h->host_id=%d", 
            __func__, sch->host_id, cam_h->host_id);
        if (sch->host_id == cam_h->host_id) {
            //shouldn't happen
            ret = -KDP_FRAMEWORK_ERRNO_RESBUSY;
            goto edevreg;
        }
    }

    framework_drv_set_drvdata(cam_h->pin_obj.pin_ctx, &cam_h->pin_obj);

    kdp_list_add_tail(&hosts, &cam_h->list);

    {
        struct sys_camera_device *cam_d;
        struct kdp_list *head = &devices;

        ret = -KDP_FRAMEWORK_ERRNO_PROBE;
        for (cam_d = kdp_list_first_entry(head, typeof(*cam_d), list);
             &cam_d->list != head;
             cam_d = kdp_list_next_entry(cam_d, list))
        {
            DSG("[%s] cam_d->connect_to=%d cam_h->host_id=%d", 
                __func__, cam_d->connect_to, cam_h->host_id);
            if (cam_d->connect_to == cam_h->host_id) {
                if ((0 == _sys_camera_probe(cam_h, cam_d)) &&
                    (0 == _sys_camera_probe_finish(cam_d))) {
                    ret = 0;
                    break;
                }
            }
        }     
    }
edevreg:
    return ret;
}

void sys_camera_host_unregister(struct sys_camera_host *cam_h)
{
    framework_drv_set_drvdata(cam_h->pin_obj.pin_ctx, NULL);
    cam_h->pin_obj.pin_ctx = NULL;
}

static int _sys_camera_power_on(struct sys_camera_device *cam_d)
{
    struct v2k_subdev *sd = sys_camera_to_subdev(cam_d);
    int ret;

    ret = v2k_subdev_call(sd, s_power, 1);
    if (ret < 0 && ret != -KDP_FRAMEWORK_ERRNO_IOCTL && ret != -KDP_FRAMEWORK_ERRNO_CTX)
        return ret;

    return 0;
}

static int _sys_camera_power_off(struct sys_camera_device *cam_d)
{
    struct v2k_subdev *sd = sys_camera_to_subdev(cam_d);
    int ret;

    ret = v2k_subdev_call(sd, s_power, 0);
    if (ret < 0 && ret != -KDP_FRAMEWORK_ERRNO_IOCTL && ret != -KDP_FRAMEWORK_ERRNO_CTX)
        return ret;

    return 0;
}

static int _sys_camera_set_fmt(struct sys_camera_device *cam_d,
        struct v2k_format *format)
{
    struct sys_camera_host *cam_h = to_sys_camera_host(cam_d->parent_pin);
    struct core_device *dev_parent = to_core_device(cam_d->parent_pin);
    struct v2k_format *fmt = format;
    int ret;
    
    DSG("[%s] - dev_parent=%s", __func__, dev_parent->name);

    ret = cam_h->ops->set_fmt(cam_d, fmt);
    if (ret < 0) {
        return ret;
    } else if (!cam_d->current_fmt || 
                cam_d->current_fmt->host_fmt->fourcc != fmt->pixelformat) {
//		DSG("Host driver hasn't set up current format correctly");
        return -KDP_FRAMEWORK_ERRNO_INVALA;
    }
    //DSG("%s width=%d height=%d", __func__, fmt->width, fmt->height);
    cam_d->user_width     = fmt->width;
    cam_d->user_height    = fmt->height;
    //cam_d->bytesperline = fmt->bytesperline;
    //cam_d->sizeimage    = fmt->sizeimage;
    cam_d->bytesperline   = sys_camera_bytes_per_line(fmt->pixelformat, cam_d->user_width);
    cam_d->sizeimage      = cam_d->bytesperline * cam_d->user_height;
    cam_d->colorspace     = (enum v2k_colorspace)fmt->colorspace;
    cam_d->field          = (enum v2k_field)fmt->field;

    /* set physical bus parameters */
    return cam_h->ops->set_bus_param(cam_d);
}
    
static int _sys_camera_clock_start(struct sys_camera_host *cam_h)
{
    int ret;

    if (!cam_h->ops->clock_start)
        return 0;

    ret = cam_h->ops->clock_start(cam_h);

    return ret;
}

static void _sys_camera_clock_stop(struct sys_camera_host *cam_h)
{
    if (!cam_h->ops->clock_stop)
        return;

    cam_h->ops->clock_stop(cam_h);

    return;
}

static int _sys_camera_reset(
        struct sys_camera_host *cam_h, struct sys_camera_device *cam_d)
{
    int ret;    
    
    if (!cam_h->ops->reset)
        return 0;

    ret = cam_h->ops->reset(cam_h);
    
    return ret;
}
const struct sys_camera_format_xlate *sys_camera_xlate_by_fourcc(
        struct sys_camera_device *cam_d, unsigned int fourcc)
{
    unsigned int i;
    for (i = 0; i < cam_d->num_user_formats; i++) {
        if (cam_d->user_formats[i].host_fmt->fourcc == fourcc)
            return cam_d->user_formats + i;
    }
    return NULL;
}

static int _sys_camera_g_fmt_vid_cap(struct v2k_dev_handle *handle, 
        void *priv, struct v2k_format *format)
{
    struct sys_camera_device *cam_d = handle->private_data;
    struct v2k_format *fmt = format;
    
	//DSG("[%s]", __func__);

    fmt->width          = cam_d->user_width;
    fmt->height         = cam_d->user_height;
    fmt->bytesperline   = cam_d->bytesperline;
    fmt->sizeimage      = cam_d->sizeimage;
    fmt->field          = cam_d->field;
    fmt->pixelformat    = cam_d->current_fmt->host_fmt->fourcc;
    fmt->colorspace     = cam_d->colorspace;

    return 0;
}

static int _sys_camera_s_fmt_vid_cap(struct v2k_dev_handle *handle, 
        void *priv, struct v2k_format *f)
{
    struct sys_camera_device *cam_d = handle->private_data;
    int ret;

    DSG("[%s]", __func__);
    
    if (priv != handle->private_data)
        return -KDP_FRAMEWORK_ERRNO_INVALA;

    ret = _sys_camera_set_fmt(cam_d, f);

    return ret;
}

static int _sys_camera_reqbufs(struct v2k_dev_handle *handle, 
        void *priv, struct v2k_requestbuffers *p)
{
    struct sys_camera_device *cam_d = handle->private_data;

    return videobuf_reqbufs(&cam_d->vb_q, p);
}

static int _sys_camera_querybuf(struct v2k_dev_handle *handle, 
        void *priv, struct v2k_buffer *p)
{
    struct sys_camera_device *cam_d = handle->private_data;

    return videobuf_querybuf(&cam_d->vb_q, p);
}

static int _sys_camera_qbuf(struct v2k_dev_handle *handle, 
        void *priv, struct v2k_buffer *p)
{
    struct sys_camera_device *cam_d = handle->private_data;
    //DSG("_sys_camera_qbuf priv=%p handle->private_data=%p", priv, handle->private_data);
    if (priv != handle->private_data)
        return -KDP_FRAMEWORK_ERRNO_INVALA;

    return videobuf_qbuf(&cam_d->vb_q, p);
}

static int _sys_camera_dqbuf(struct v2k_dev_handle *handle, 
        void *priv, struct v2k_buffer *p)
{
    struct sys_camera_device *cam_d = handle->private_data;
    //DSG("_sys_camera_dqbuf priv=%p handle->private_data=%p", priv, handle->private_data);
    if (priv != handle->private_data)
        return -KDP_FRAMEWORK_ERRNO_INVALA;

    return videobuf_dqbuf(&cam_d->vb_q, p, false);
}


static int sys_camera_init_user_formats(struct sys_camera_device *cam_d)
{
    struct v2k_subdev *sd = sys_camera_to_subdev(cam_d);
    struct sys_camera_host *cam_h = to_sys_camera_host(cam_d->parent_pin);
    unsigned int i, fmts = 0;
    unsigned int fourcc;

    DSG("[%s]", __func__);

    while (!v2k_subdev_call(sd, enum_fmt, fmts, &fourcc))
        fmts++;

    if (!fmts)
        return -KDP_FRAMEWORK_ERRNO_IO;	

    cam_d->user_formats = calloc(fmts, sizeof(struct sys_camera_format_xlate));
    if (!cam_d->user_formats)
        return -KDP_FRAMEWORK_ERRNO_NOMEM;

    //DSG("<%s> Found %d supported formats.", __func__, fmts);

    for (i = 0; i < fmts; i++) {
        v2k_subdev_call(sd, enum_fmt, i, &fourcc);
        cam_d->user_formats[i].host_fmt = sys_camera_get_fmtdesc(fourcc);
    }
    cam_d->num_user_formats = fmts;
    cam_d->current_fmt = &cam_d->user_formats[0];

    DSG("[%s] done", __func__);

    return 0;
}

static void _sys_camera_free_user_formats(struct sys_camera_device *cam_d)
{
    struct sys_camera_host *cam_h = to_sys_camera_host(cam_d->parent_pin);

    cam_d->current_fmt = NULL;
    cam_d->num_user_formats = 0;
    if (cam_d->user_formats)
        free(cam_d->user_formats);
    cam_d->user_formats = NULL;
}

static void _sys_camera_remove_device(struct sys_camera_device *cam_d)
{
    struct sys_camera_host *cam_h = to_sys_camera_host(cam_d->parent_pin);

    _sys_camera_clock_stop(cam_h);

    cam_h->cam_d = NULL;
}


static int _sys_camera_open(struct v2k_dev_handle *handle)
{
    struct video_device_context *vdc = video_devdata(handle); 
    struct sys_camera_device *cam_d;
    struct sys_camera_host *cam_h;
    int ret;

    DSG("[%s]", __func__);

    if (!vdc) {
        return -KDP_FRAMEWORK_ERRNO_CTX;
    }

    cam_d = framework_drv_get_drvdata(&vdc->pin_ctx);
    cam_h = to_sys_camera_host(cam_d->parent_pin);

    if (!to_sys_camera_control(cam_d)) {
        /* No device driver attached */
        ret = -KDP_FRAMEWORK_ERRNO_CTX;
        goto econtrol;
    }

    {
        /* turn on the camera sensor */
        ret = _sys_camera_power_on(cam_d);
        if (ret < 0)
            goto epower;

//        struct sys_camera_desc *cam_desc = to_sys_camera_desc(cam_d);
//        if (cam_desc->subdev_desc.reset)
//                cam_desc->subdev_desc.reset(cam_d->child_pin);

        /* start controller clock */
        ret = _sys_camera_clock_start(cam_h);
        if (ret < 0) {
            // DSG("Couldn't activate the camera: %d", ret);
            goto eclkstart;
        }
        //james2();
        /* try to reset controller & camera sensor */
        ret = _sys_camera_reset(cam_h, cam_d);

        ret = cam_h->ops->init_videobuf(&cam_d->vb_q, cam_d);
        if (ret < 0)
            goto einitvb;

    }

    handle->private_data = cam_d;
    //DSG("[%s] done", __func__);
    return 0;

einitvb:

    _sys_camera_power_off(cam_d);
epower:
    _sys_camera_remove_device(cam_d);
eclkstart:

econtrol:

    return ret;
}

static int _sys_camera_close(struct v2k_dev_handle *handle)
{
    return 0;
}

static int _sys_camera_querycap(struct v2k_dev_handle *handle, 
        void  *priv, struct v2k_capability *cap)
{
    struct sys_camera_device *cam_d = handle->private_data;
    struct sys_camera_host *cam_h = to_sys_camera_host(cam_d->parent_pin);
    struct core_device *dev_parent = to_core_device(cam_d->parent_pin);

    DSG("[%s] dev_parent=%s", __func__, dev_parent->name);

    strncpy(cap->driver, dev_parent->name, sizeof(cap->driver) - 1);
    //strncpy(cap->driver, cam_h->drv_name, sizeof(cap->driver) - 1);
    //cap->driver[strlen(cap->driver) - 1] = '\0';

    int ret = cam_h->ops->querycap(cam_h, cap);

    return ret;
}

static int _sys_camera_streamon(struct v2k_dev_handle *handle, 
        void *priv, unsigned long arg)
{
    struct sys_camera_device *cam_d = handle->private_data;
    struct sys_camera_host *cam_h = to_sys_camera_host(cam_d->parent_pin);
    struct v2k_subdev *sd = sys_camera_to_subdev(cam_d);
    
    int ret;
    
    ret = videobuf_streamon(&cam_d->vb_q, arg);
    if (!ret)
        v2k_subdev_call(sd, s_stream, 1);

    return ret;
}

static int _sys_camera_streamoff(struct v2k_dev_handle *handle, 
        void *priv, unsigned long arg)
{
    struct sys_camera_device *cam_d = handle->private_data;
    struct sys_camera_host *cam_h = to_sys_camera_host(cam_d->parent_pin);    
    struct v2k_subdev *sd = sys_camera_to_subdev(cam_d);
//    int ret;

//    if (cam_d->streamer != handle)
//        return -KDP_FRAMEWORK_ERRNO_RESBUSY;

    videobuf_streamoff(&cam_d->vb_q, arg);

//    v2k_subdev_call(sd, s_stream, 0);

    return 0;//ret;
}

static int sys_camera_i2c_init(struct sys_camera_device *cam_d,
        struct sys_camera_desc *sdesc)
{
    int ret = -KDP_FRAMEWORK_ERRNO_CTX;
    struct sensor_device *dev;
    //struct i2c_adapter *adap;
    struct v2k_subdev *sd;
    struct sys_camera_desc *cam_desc = sdesc;
    struct sys_camera_info *info = cam_desc->info;

    
    //adap = i2c_get_adapter(cam_desc->i2c_adapter_id); 
    //DSG("[%s] cam_desc_h->i2c_adapter_id=%d adap=%p", 
    //    __func__, cam_desc->i2c_adapter_id, adap);

    dev = (struct sensor_device *)calloc(1, sizeof(*dev));
    if (dev) {
        struct sensor_driver *sensor_drv;
        //dev->adapter = adap;
        dev->pin_ctx.platform_data = NULL;
        dev->addr = info->slave_addr;
        //DSG("info->name <%s>", info->name);

        sensor_drv = to_sensor_driver(find_driver(info->name));
        DSG("[%s] sensor_drv=%p", __func__, sensor_drv);
    #ifndef SYS_CAMERA_BUG_0720 // 
        ret = 0;
        if (sensor_drv)
    #else        
        if (sensor_drv->probe)
            ret = sensor_drv->probe(dev);
        DSG("[%s] sensor_drv->probe ret=%d", __func__, ret);
        if (0 == ret)
    #endif        
        {
            sd = framework_drv_get_drvdata(&dev->pin_ctx);
            DSG("[%s] sd=%p", __func__, sd);
            if (sd) {
                dev = v2k_get_subdevdata(sd);
                cam_d->child_pin = &dev->pin_ctx;
            }
        }
        else {
            free(dev);
            ret = -KDP_FRAMEWORK_ERRNO_CTX;
        } 
    }

    return ret;
}

static int _sys_camera_device_register(struct sys_camera_device *cam_d)
{
//    struct sys_camera_device *scd;
    int num = 0;

    cam_d->idx_on_host = num;

    kdp_list_add_tail(&devices, &cam_d->list);

    return 0;
}

static const struct v2k_dev_operations sys_camera_dev_ops = {
    .open       = _sys_camera_open,
    .release    = _sys_camera_close,
    .ioctl      = video_default_ioctl,
};

static const struct v2k_ioctl_ops sys_camera_ioctl_ops = {
    .vioc_querycap      = _sys_camera_querycap,
    .vioc_g_fmt_vid_cap = _sys_camera_g_fmt_vid_cap,
    .vioc_s_fmt_vid_cap = _sys_camera_s_fmt_vid_cap,
    .vioc_reqbufs       = _sys_camera_reqbufs,
    .vioc_querybuf      = _sys_camera_querybuf,
    .vioc_qbuf          = _sys_camera_qbuf,
    .vioc_dqbuf         = _sys_camera_dqbuf,
    .vioc_streamon      = _sys_camera_streamon,
    .vioc_streamoff     = _sys_camera_streamoff,
};

static int _sys_camera_probe(
        struct sys_camera_host *cam_h, struct sys_camera_device *cam_d)
{
    int ret;
    struct sys_camera_desc *cam_desc = to_sys_camera_desc(cam_d);
    struct video_device_context *vdc;

    DSG("[%s]", __func__);
    
    vdc = calloc(1, sizeof(struct video_device_context));
    if (!vdc) {
        ret = -KDP_FRAMEWORK_ERRNO_NOMEM;
        goto evdc;
    }

    vdc->dev_ops   = &sys_camera_dev_ops;
    vdc->ioctl_ops = &sys_camera_ioctl_ops;
    cam_d->parent_pin = cam_h->pin_obj.pin_ctx; 
    cam_d->vdc = vdc;
    //DSG("[%s] cam_desc->info->name=%s", cam_desc->info->name);
    if (cam_desc->info) {
        ret = sys_camera_i2c_init(cam_d, cam_desc);
        if (ret < 0)
            goto eadd;
    }
    DSG("[%s] sys_camera_i2c_init -- O", __func__);
    return 0;

eadd:
    if (cam_d->vdc) {
        video_device_release(cam_d->vdc);
        cam_d->vdc = NULL;
    }
evdc:
    DSG("[%s] sys_camera_i2c_init -- X, ret=%x", __func__, ret);
    return ret;
}

static int _sys_camera_probe_finish(struct sys_camera_device *cam_d)
{
    int ret;
    struct sys_camera_host *cam_h = to_sys_camera_host(cam_d->parent_pin);
    struct v2k_subdev *sd = sys_camera_to_subdev(cam_d); //kdp520_i2c
    
    DSG("[%s]", __func__);
    
    ret = sys_camera_init_user_formats(cam_d);
    if (0 == ret) {
        cam_d->field = V2K_FIELD_ANY;

        framework_drv_set_drvdata(&cam_d->vdc->pin_ctx, cam_d);
        ret = video_device_register(cam_d->vdc);//, cam_d->connect_to);
        if (ret < 0) {
            _sys_camera_free_user_formats(cam_d);
            _sys_camera_remove_device(cam_d);
        }
    }
    else {
        _sys_camera_remove_device(cam_d);
    }

    return ret;
}

static int _sys_camera_core_probe(struct core_device *core_d) 
{
    struct sys_camera_desc *cam_desc = (struct sys_camera_desc *)core_d->pin_ctx.platform_data;
    struct sys_camera_device *cam_d;

    DSG("[%s] %s", __func__, core_d->name);

    cam_d = (struct sys_camera_device *)calloc(1, sizeof(*cam_d));
    if (!cam_d)
        return -KDP_FRAMEWORK_ERRNO_NOMEM;

    cam_d->connect_to = cam_desc->connect_to; //sensor_link : connect_to
    cam_d->sdesc = cam_desc;

    framework_drv_set_drvdata(&core_d->pin_ctx, (void*)cam_d);

    cam_d->user_width = VGA_LANDSCAPE_WIDTH;
    cam_d->user_height = VGA_LANDSCAPE_HEIGHT;
    
    return _sys_camera_device_register(cam_d);
}

static int _sys_camera_core_remove(struct core_device *core_d)
{
    struct sys_camera_device *cam_d = framework_drv_get_drvdata(&core_d->pin_ctx); 

    int i;

    if (!cam_d)
        return -KDP_FRAMEWORK_ERRNO_INVALA;
    
    i = core_d->uuid;
    if (i < 0)
        i = 0;

    //clear_bit(i, device_map);
    kdp_list_del(&cam_d->list);

    return 0;
}

void sys_camera_global_init(void)
{   
    // This should be located at driver layer ...
    //SCU_EXTREG_CLK_MUX_SEL_SET_csirx1_clk(1);
    //0717
   SCU_EXTREG_CLK_EN1_SET_csirx0_TxEscClk(1);
   SCU_EXTREG_CLK_EN1_SET_csirx0_vc0(1);
   SCU_EXTREG_CLK_EN1_SET_csirx0_csi(1);        
   
   SCU_EXTREG_CLK_EN1_SET_csirx1_TxEscClk(1);
   SCU_EXTREG_CLK_EN1_SET_csirx1_vc0(1);
   SCU_EXTREG_CLK_EN1_SET_csirx1_csi(1);        
   
   masked_outw( SCU_EXTREG_CSIRX_CTRL0, 
               ((SCU_EXTREG_CSIRX_CTRL0_ClkLnEn | SCU_EXTREG_CSIRX_CTRL0_Enable) |
                (SCU_EXTREG_CSIRX_CTRL0_apb_rst_n | SCU_EXTREG_CSIRX_CTRL0_pwr_rst_n | 
                 SCU_EXTREG_CSIRX_CTRL0_sys_rst_n)),
               ((SCU_EXTREG_CSIRX_CTRL0_ClkLnEn | SCU_EXTREG_CSIRX_CTRL0_Enable) |
                (SCU_EXTREG_CSIRX_CTRL0_apb_rst_n | SCU_EXTREG_CSIRX_CTRL0_pwr_rst_n | 
                 SCU_EXTREG_CSIRX_CTRL0_sys_rst_n)));
   
   masked_outw( SCU_EXTREG_CSIRX_CTRL1, 
               ((SCU_EXTREG_CSIRX_CTRL1_ClkLnEn | SCU_EXTREG_CSIRX_CTRL1_Enable) |
                (SCU_EXTREG_CSIRX_CTRL1_apb_rst_n | SCU_EXTREG_CSIRX_CTRL1_pwr_rst_n | 
                 SCU_EXTREG_CSIRX_CTRL1_sys_rst_n)),
               ((SCU_EXTREG_CSIRX_CTRL1_ClkLnEn | SCU_EXTREG_CSIRX_CTRL1_Enable) |
                (SCU_EXTREG_CSIRX_CTRL1_apb_rst_n | SCU_EXTREG_CSIRX_CTRL1_pwr_rst_n | 
                 SCU_EXTREG_CSIRX_CTRL1_sys_rst_n)));    

   masked_outw( SCU_EXTREG_DPI2AHB_CTRL, 
               (SCU_EXTREG_DPI2AHB_CTRL_rst_n | SCU_EXTREG_DPI2AHB_CTRL_apb_rst_n |
                SCU_EXTREG_DPI2AHB_CTRL_apb_rst_n | SCU_EXTREG_DPI2AHB_CTRL_pwr_rst_n), 
               (SCU_EXTREG_DPI2AHB_CTRL_rst_n | SCU_EXTREG_DPI2AHB_CTRL_apb_rst_n |
                SCU_EXTREG_DPI2AHB_CTRL_apb_rst_n | SCU_EXTREG_DPI2AHB_CTRL_pwr_rst_n));
                   
   masked_outw( SCU_EXTREG_DPI2AHB_CTRL, 
               (SCU_EXTREG_DPI2AHB_CTRL_rst_n_1 | SCU_EXTREG_DPI2AHB_CTRL_apb_rst_n_1 |
                SCU_EXTREG_DPI2AHB_CTRL_sys_rst_n_1 | SCU_EXTREG_DPI2AHB_CTRL_pwr_rst_n_1),
               (SCU_EXTREG_DPI2AHB_CTRL_rst_n_1 | SCU_EXTREG_DPI2AHB_CTRL_apb_rst_n_1 |
                SCU_EXTREG_DPI2AHB_CTRL_sys_rst_n_1 | SCU_EXTREG_DPI2AHB_CTRL_pwr_rst_n_1));    
}

extern struct core_device kdp520_camera_dev;
static struct core_driver sys_camera_core = {
    .probe      = _sys_camera_core_probe,
    .remove     = _sys_camera_core_remove,
    .driver = {
        .name   = "sys_camera_core",
    },
    .core_dev   = &kdp520_camera_dev,
};
KDP_CORE_DRIVER_SETUP(sys_camera_core, 4);

#if (V2K_ENABLE_TYPE == V2K_ENABLE_HMX2056_OV9286) || \
    (V2K_ENABLE_TYPE == V2K_ENABLE_GC2145_SC132GS) || \
    (V2K_ENABLE_TYPE == V2K_ENABLE_SC132GS_GC2145)
extern struct core_device kdp520_camera_dev_1;
static struct core_driver sys_camera_core_1 = {
    .probe      = _sys_camera_core_probe,
    .remove     = _sys_camera_core_remove,
    .driver = {
        .name   = "sys_camera_core_1",
    },
    .core_dev   = &kdp520_camera_dev_1,
};
KDP_CORE_DRIVER_SETUP(sys_camera_core_1, 4);
#endif

#endif
#endif

