/* Copyright (c) 2020 Kneron, Inc. All Rights Reserved.
*
* The information contained herein is property of Kneron, Inc.
* Terms and conditions of usage are described in detail in Kneron
* STANDARD SOFTWARE LICENSE AGREEMENT.
*
* Licensees are granted free, non-transferable use of the information.
* NO WARRANTY of ANY KIND is provided. This heading must NOT be removed
* from the file.
*/

/******************************************************************************
*  Filename:
*  ---------
*  kdrv_adc.c
*
*  Project:
*  --------
*  KL520
*
*  Description:
*  ------------
*  This ADC driver is for Generic Analog-to-Digital Converter
*  HW: Faraday FTADC010
*
*  Author:
*  -------
*  Teresa Chen
**
******************************************************************************/

/******************************************************************************
Head Block of The File
******************************************************************************/

#ifdef USE_KDRV
#include <stdio.h>
#include <string.h>
#include "io.h"
#include "kneron_mozart.h"
#include "kdrv_adc.h"
#include "kmdw_console.h"

#define NUM_TDC         4
#define SCANMODE        SCANMODE_CONT

kdrv_adc_resource_t kdrv_adc_resource = {
    ADC_FTTSC010_0_PA_BASE,
    ADC_FTADCC010_IRQ,
};

void _kdrv_adc_config(void)
{
    struct kdrv_adc_regs_t *regs = (struct kdrv_adc_regs_t *)kdrv_adc_resource.io_base;

    uint32_t val;

    /* set ADC clock pre-scaler control register (offset 0x11C) */
    //outw(regs->prescal, val);

    /* set ADC timing parameter register (offset 0x110) */
    //outw(regs->tparam, val);

    /* set ADC sampling timing parameter control register (offset 0x114) */
    //outw(regs->smpr, val);

    /* set ADC scan sequence register (offset 0x120) */
    //outw(regs->sqr, val);

    /* set ADC interrupt enable register (offset 0x108) */
    //outw(regs->inten, val);

    /* set ADC control register (offset 0x100) */
    val = 0;
    val = ADC_EN | SWSTART | SCANMODE_CONT | SCAN_NUM(4);
    //dbg_msg_console("val = 0x%x", val);
    outw(&regs->ctrl, val);
    //_print_adc_register(&kdrv_adc_resource);
}

kdrv_status_t kdrv_adc_initialize(void)
{
    uint32_t i;
    uint32_t val = 0;
    //char Ch;
    struct kdrv_adc_regs_t *regs = (struct kdrv_adc_regs_t *)kdrv_adc_resource.io_base;

    //Enable ADCCLK in SCU base 0xC2380000
    //Set the 22th bit of the "Clock Enable Register1"(offset = 0x018) to 1
    val = inw(SCU_EXTREG_PA_BASE + 0x018);
    outw(SCU_EXTREG_PA_BASE + 0x018, (val | (1 << 22 )));

    osDelay(1);
    _kdrv_adc_config();

    // Wait for scanning done.
    for(i = 0; i < NUM_TDC; i++) {
        while(!(readl(&regs->intst) & CH_INTRSTS(i))) {}
    }

	while(!(readl(&regs->intst) & ADC_DONE_INTSTS)){};
    /*while(1)
    {
        dbg_msg_console("AIN0: %d \r\nAIN1: %d \r\nAIN2: %d \r\nAIN3: %d \r\n ===\r\n",regs->data[0],regs->data[1],regs->data[2],regs->data[3]);
        dbg_msg_console("press anykey to measureing, and 'q' to exit");
        //_print_adc_register(&kdrv_adc_resource);
        Ch = kmdw_console_getc();
        if (Ch == 'q')
            break;
    }*/
    return KDRV_STATUS_OK;
}

kdrv_status_t kdrv_adc_uninitialize(kdrv_adc_resource_t *res)
{
    return KDRV_STATUS_OK;
}

kdrv_status_t kdrv_adc_rest(kdrv_adc_resource_t *res)
{
    int val;
    struct kdrv_adc_regs_t *regs = (struct kdrv_adc_regs_t *)res->io_base;

    //disable
    val = readl(&regs->ctrl);
    val &= ~ADC_EN; //clear enable bit
    writel(val, &regs->ctrl);

    //enable
    val |= ADC_EN; //clear enable bit
    writel(val, &regs->ctrl);

    return KDRV_STATUS_OK;
}

kdrv_status_t kdrv_adc_enable(kdrv_adc_resource_t *res, int mode)
{
    int val;
    struct kdrv_adc_regs_t *regs = (struct kdrv_adc_regs_t *)res->io_base;

    val = readl(&regs->ctrl);
    val &= ~ADC_EN; //clear enable bit
    val |= mode;
    writel(val, &regs->ctrl);

    return KDRV_STATUS_OK;
}

static void kdrv_adc_force_update_temp(kdrv_adc_resource_t *res)
{
    kdrv_adc_enable(res, ADC_EN);
}

int kdrv_adc_read(int id)
{
    uint32_t val = 0;
    struct kdrv_adc_regs_t *regs = (struct kdrv_adc_regs_t *)kdrv_adc_resource.io_base;

    if(SCANMODE == SCANMODE_SGL)
        kdrv_adc_force_update_temp(&kdrv_adc_resource);

    val = readl(&regs->data[id]);

    DSG("print raw data: %d\n", val);
    
    return val;
}

void _print_adc_register(kdrv_adc_resource_t *res)
{
    uint32_t val = 0;
    struct kdrv_adc_regs_t *regs = (struct kdrv_adc_regs_t *)res->io_base;

    val = inw(&regs->prescal);  //0x11C
    DSG("0x%x = 0x%x", &regs->prescal, val);
    val = inw(&regs->tparam);   //0x110
    DSG("0x%x = 0x%x", &regs->tparam, val);
    val = inw(&regs->smpr);     //0x114
    DSG("0x%x = 0x%x", &regs->smpr, val);
    val = inw(&regs->sqr);          //0x120
    DSG("0x%x = 0x%x", &regs->sqr, val);
    val = inw(&regs->inten);        //0x108
    DSG("0x%x = 0x%x", &regs->inten, val);
    val = inw(&regs->ctrl);     //0x100
    DSG("0x%x = 0x%x", &regs->ctrl, val);

}
#endif
