/* Copyright (c) 2020 Kneron, Inc. All Rights Reserved.
 *
 * The information contained herein is property of Kneron, Inc.
 * Terms and conditions of usage are described in detail in Kneron
 * STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information.
 * NO WARRANTY of ANY KIND is provided. This heading must NOT be removed
 * from the file.
 */

/******************************************************************************
*  Filename:
*  ---------
*  kdrv_fb_mgr.c
*
*  Project:
*  --------
*  KL520
*
*  Description:
*  ------------
*  Configuration for KL520
*
*  Author:
*  -------
*  XXX

**
******************************************************************************/

/******************************************************************************
Head Block of The File
******************************************************************************/

#ifdef USE_KDRV
#include <string.h>
#include "board_kl520.h"
#include "kmdw_memory.h"
#include "kdrv_fb_mgr.h"
#include "kmdw_console.h"
#include "framework/irqflags.h"
#include "RTX_Config.h"

#define USE_GUARDBAND           4   // 1/n on both sides

#define INF_WAIT_MAX_TIME_MS    500
#define INF_WAIT_OS_DELAY_MS    5
#define INF_WAIT_OS_DELAY_TICKS (OS_TICK_FREQ * INF_WAIT_OS_DELAY_MS / 1000)

//#define FB_DEBUG

#ifdef FB_DEBUG
#define fb_msg(fmt, ...) err_msg("[%8.8d] " fmt, osKernelGetTickCount(), ##__VA_ARGS__)
#else
#define fb_msg(fmt, ...)
#endif

static struct kdrv_fb_mgr_s {
    int         buf_size;
    int         buf_num;

    uint32_t    buf_addr[MAX_FRAME_BUFFER];
    uint32_t    buf_seq_num[MAX_FRAME_BUFFER];

    int         frame_info_size;
    uint32_t    frame_info_buf; // addr of [MAX_FRAME_BUFFER][frame_info_size]

    int         write_idx;      // camera
    int         write_next_idx; // camera
    int         write_done_idx; // camera
    uint32_t    write_count;

    int         inf_idx;        // inference
    int         inf_done_idx;   // inference
    uint32_t    inf_count;

    int         read_idx;       // drawing
    int         read_done_idx;  // drawing
    uint32_t    read_count;

    int         disp_idx;       // display
    int         disp_src;

    fb_write_done_notify    write_done_cb;
    bool        init_done;

} kdrv_frame_buffer[IMGSRC_NUM];

static struct kdrv_fb_mgr_s *fb_p[IMGSRC_NUM] = {
    &kdrv_frame_buffer[0],
#if (IMGSRC_IN_1 == YES)
    &kdrv_frame_buffer[1],
#endif
};

int kdrv_fb_mgr_init(int cam_idx, uint32_t buf_size, int buf_num, int frame_info_size)
{
    struct kdrv_fb_mgr_s *fbp = fb_p[cam_idx];
    int i, size;

    if(true == fbp->init_done)
        return 0;

#if USE_GUARDBAND
    size = buf_size + (2 * buf_size / USE_GUARDBAND);
#else
    size = buf_size;
#endif

    if ((buf_num > MAX_FRAME_BUFFER) || (0 == buf_num))
        buf_num = MAX_FRAME_BUFFER;

    for (i = 0; i < buf_num; i++) {
        fbp->buf_addr[i] = kmdw_ddr_reserve(size);
#if USE_GUARDBAND
        fbp->buf_addr[i] += (buf_size / USE_GUARDBAND);
#endif
        info_msg("  cam %d: frame buf[%d] : 0x%x\n", cam_idx, i, fbp->buf_addr[i]);
    }

    fbp->frame_info_size = frame_info_size;
    if (frame_info_size > 0) {
        fbp->frame_info_buf = kmdw_ddr_reserve(frame_info_size * buf_num);
        info_msg("  cam %d: frame_info buf: 0x%x, size: %d*%d\n",
                cam_idx, fbp->frame_info_buf, buf_num, frame_info_size);
    }

    fbp->write_idx = fbp->write_next_idx = fbp->write_done_idx = -1;
    fbp->read_idx = fbp->read_done_idx = -1;
    fbp->inf_idx = fbp->inf_done_idx = -1;
    fbp->disp_idx = -1;

    fbp->buf_size = buf_size;
    fbp->buf_num = buf_num;
    fbp->init_done = true;
    return 0;
}

uint32_t kdrv_fb_mgr_next_write(int cam_idx, int *write_idx)
{
    struct kdrv_fb_mgr_s *fbp = fb_p[cam_idx];
    uint32_t temp;

    local_irq_disable();

    if (fbp->write_idx == -1)
        fbp->write_idx = fbp->write_next_idx;

    fb_msg("[%d]%s: w-next:%d (disp:%d, r-done:%d, r:%d, inf-done:%d, inf:%d, w-done:%d)\n",
        cam_idx, __func__, fbp->write_next_idx,
        fbp->disp_idx, fbp->read_done_idx, fbp->read_idx,
        fbp->inf_done_idx, fbp->inf_idx, fbp->write_done_idx);

    // skip the buffers being or could be used for inference/read
    temp = fbp->write_next_idx;
    do {
        temp = (temp + 1) % fbp->buf_num;
    } while (temp == fbp->disp_idx || temp == fbp->read_done_idx ||
             temp == fbp->read_idx || temp == fbp->inf_done_idx ||
             temp == fbp->inf_idx || temp == fbp->write_done_idx);

    fbp->write_next_idx = temp;

    fb_msg("[%d]%s: w-next:%d (final)\n",
        cam_idx, __func__, fbp->write_next_idx);

    if (write_idx)
        *write_idx = fbp->write_next_idx;

    local_irq_enable();

    return fbp->buf_addr[fbp->write_next_idx];
}

int kdrv_fb_mgr_write_done(int cam_idx, int write_idx)
{
    struct kdrv_fb_mgr_s *fbp = fb_p[cam_idx];

    local_irq_disable();

    fb_msg("[%d]%s: %d\n", cam_idx, __func__, write_idx);

    if (fbp->inf_idx != -1 ) {
        if (fbp->write_next_idx == fbp->read_idx) {
            err_msg("\n [%d]%s: [%d done] w-next:%d == read_idx (prev w-done:%d, inf:%d, inf-done:%d) !\n",
                cam_idx, __func__, write_idx, fbp->write_next_idx, fbp->write_done_idx,
                fbp->inf_idx, fbp->inf_done_idx);
        } else if (fbp->write_next_idx == fbp->inf_idx) {
            err_msg("\n [%d]%s: [%d done] w-next:%d == inf_idx %d !!\n",
                cam_idx, __func__, write_idx, fbp->write_next_idx, fbp->inf_idx);
        }
    }

    fbp->write_done_idx = write_idx;
    fbp->write_idx = fbp->write_next_idx;

    /* Mark with sequential number */
    fbp->write_count++;
    fbp->buf_seq_num[write_idx] = fbp->write_count;

    if (fbp->write_done_cb)
        fbp->write_done_cb(cam_idx, write_idx);

    local_irq_enable();

    return 0;
}

void kdrv_fb_mgr_free_write_buf(int cam_idx)
{
    struct kdrv_fb_mgr_s *fbp = fb_p[cam_idx];
    int i;

    local_irq_disable();

    fbp->write_idx = -1;
    fbp->write_next_idx = -1;
    fbp->write_done_idx = -1;

    for (i = 0; i < fbp->buf_num; i++) {
        fbp->buf_seq_num[i] = 0;
    }

    fb_msg("[%d]%s\n", cam_idx, __func__);

    local_irq_enable();
}

uint32_t kdrv_fb_mgr_get_buf_seq_num(int cam_idx, int idx)
{
    struct kdrv_fb_mgr_s *fbp = fb_p[cam_idx];

    return fbp->buf_seq_num[idx];
}

uint32_t kdrv_fb_mgr_get_current_buf_seq_num(int cam_idx)
{
    struct kdrv_fb_mgr_s *fbp = fb_p[cam_idx];

    return fbp->write_count;
}

uint32_t kdrv_fb_mgr_get_buf(int cam_idx, int idx)
{
    struct kdrv_fb_mgr_s *fbp = fb_p[cam_idx];

    local_irq_disable();

    idx %= fbp->buf_num;
    uint32_t buf_addr = fbp->buf_addr[idx];
    memset((void *)buf_addr, 0, fbp->buf_size);

    fb_msg("[%d]%s: %d\n", cam_idx, __func__, idx);

    local_irq_enable();

    return buf_addr;
}

uint32_t kdrv_fb_mgr_get_frame_info_buf(int cam_idx, int idx)
{
    struct kdrv_fb_mgr_s *fbp = fb_p[cam_idx];
    uint32_t buf_addr;

    local_irq_disable();

    if (fbp->frame_info_buf) {
        idx %= fbp->buf_num;
        buf_addr = fbp->frame_info_buf + (idx * fbp->frame_info_size);
    } else {
        buf_addr = 0;
    }

    fb_msg("[%d]%s: %d\n", cam_idx, __func__, idx);

    local_irq_enable();

    return buf_addr;
}

uint32_t kdrv_fb_mgr_next_read(int cam_idx, int *read_idx)
{
    struct kdrv_fb_mgr_s *fbp = fb_p[cam_idx];
    int try_max_count = INF_WAIT_MAX_TIME_MS / INF_WAIT_OS_DELAY_MS;

next_read_buf:
    local_irq_disable();

    if (fbp->write_done_idx != -1) {
        fbp->read_idx = fbp->write_done_idx;
        fb_msg("[%d]%s: %d\n", cam_idx, __func__, fbp->read_idx);
    } else {
        // let's wait for next available one
        if (try_max_count-- >= 0) {
            local_irq_enable();
            osDelay(INF_WAIT_OS_DELAY_TICKS);
            goto next_read_buf;
        }

        if (fbp->read_idx == -1)
            fbp->read_idx = 0;
        fb_msg("[%d]%s: reuse read_idx:%d after waiting %d ms\n",
                cam_idx, __func__, fbp->read_idx, INF_WAIT_MAX_TIME_MS);
    }

    if (read_idx)
        *read_idx = fbp->read_idx;

    fbp->read_count++;

    local_irq_enable();

    return fbp->buf_addr[fbp->read_idx];
}

int kdrv_fb_mgr_read_done(int cam_idx, int read_idx)
{
    struct kdrv_fb_mgr_s *fbp = fb_p[cam_idx];

    local_irq_disable();

    fbp->read_done_idx = read_idx;

    fb_msg("[%d]%s: %d\n", cam_idx, __func__, read_idx);

    local_irq_enable();

    return 0;
}

void kdrv_fb_mgr_free_read_buf(int cam_idx)
{
    struct kdrv_fb_mgr_s *fbp = fb_p[cam_idx];

    local_irq_disable();

    fbp->read_idx = -1;
    fbp->read_done_idx = -1;

    fb_msg("[%d]%s\n", cam_idx, __func__);

    local_irq_enable();
}

uint32_t kdrv_fb_mgr_next_inf(int cam_idx, int *inf_idx)
{
    struct kdrv_fb_mgr_s *fbp = fb_p[cam_idx];
    int try_max_count = INF_WAIT_MAX_TIME_MS / INF_WAIT_OS_DELAY_MS;

next_inf_buf:
    local_irq_disable();

    // inference uses the latest write_done_idx
    if (fbp->write_done_idx == fbp->inf_done_idx) {
        // latest write_done_idx is already used. let's wait for new one
        if (try_max_count-- >= 0) {
            local_irq_enable();
            osDelay(INF_WAIT_OS_DELAY_TICKS);
            goto next_inf_buf;
        }
        fb_msg("[%d]%s: reuse w-done:%d after waiting %d ms\n",
                cam_idx, __func__, fbp->write_done_idx, INF_WAIT_MAX_TIME_MS);
    } else {
        fb_msg("[%d]%s: %d\n", cam_idx, __func__, fbp->write_done_idx);
    }

    if (fbp->inf_done_idx != -1) {
        int new_buffers = fbp->write_done_idx - fbp->inf_done_idx;
        if (new_buffers < 0)
            new_buffers += fbp->buf_num;

        if (new_buffers > 1) {
            fb_msg("[%d]%s: skip %d buffers: from %d to %d\n",
                cam_idx, __func__, new_buffers - 1,
                fbp->inf_done_idx, fbp->write_done_idx);
        }
    }

    // inference uses the latest write_done_idx
    fbp->inf_idx = fbp->write_done_idx;

    if (inf_idx)
        *inf_idx = fbp->inf_idx;

    fbp->inf_count++;

    local_irq_enable();

    return fbp->buf_addr[fbp->inf_idx];
}

int kdrv_fb_mgr_inf_done(int cam_idx, int inf_idx)
{
    struct kdrv_fb_mgr_s *fbp = fb_p[cam_idx];

    local_irq_disable();

    fbp->inf_done_idx = inf_idx;

    fb_msg("[%d]%s: %d\n", cam_idx, __func__, inf_idx);

    local_irq_enable();

    return 0;
}

void kdrv_fb_mgr_free_inf_buf(int cam_idx)
{
    struct kdrv_fb_mgr_s *fbp = fb_p[cam_idx];

    local_irq_disable();

    fbp->inf_idx = -1;
    fbp->inf_done_idx = -1;

    // Free read buffers as well to save call
    fbp->read_idx = -1;
    fbp->read_done_idx = -1;

    fb_msg("[%d]%s\n", cam_idx, __func__);

    local_irq_enable();
}

uint32_t kdrv_fb_mgr_next_display(int cam_idx, int *disp_idx)
{
    struct kdrv_fb_mgr_s *fbp = fb_p[cam_idx];
    int     idx = -1;

    local_irq_disable();

    if (fbp->disp_src == DISPLAY_SRC_READ) {
        idx = fbp->read_done_idx;
        if (idx != -1)
            fb_msg("[%d]%s: %d (read-done)\n", cam_idx, __func__, idx);
    }

    if (fbp->disp_src == DISPLAY_SRC_INF || (fbp->disp_src == DISPLAY_SRC_READ && idx == -1)) {
        idx = fbp->inf_done_idx;
        if (idx != -1)
            fb_msg("[%d]%s: %d (inf-done)\n", cam_idx, __func__, idx);
    }

    if (fbp->disp_src == DISPLAY_SRC_WRITE || idx == -1) {
        idx = fbp->write_done_idx;
        if (idx != -1)
            fb_msg("[%d]%s: %d (write-done)\n", cam_idx, __func__, idx);
    }

    if (idx == -1) {
        // No new buffer. Use the last buffer in the pool and blank it
        fbp->disp_idx = fbp->buf_num - 1;
        memset((void *)fbp->buf_addr[fbp->disp_idx], 0, fbp->buf_size);
    } else {
        fbp->disp_idx = idx;
    }

    if (disp_idx)
        *disp_idx = fbp->disp_idx;

    local_irq_enable();

    return fbp->buf_addr[fbp->disp_idx];
}

int kdrv_fb_mgr_display_set_src(int cam_idx, int disp_src)
{
    struct kdrv_fb_mgr_s *fbp = fb_p[cam_idx];

    if (disp_src >= DISPLAY_SRC_READ && disp_src <= DISPLAY_SRC_WRITE) {
        fbp->disp_src = disp_src;
        return 0;
    } else {
        return -1;
    }
}

int kdrv_fb_mgr_notifier_register(int cam_idx, fb_write_done_notify callback)
{
    struct kdrv_fb_mgr_s *fbp = fb_p[cam_idx];

    fbp->write_done_cb = callback;
    return 0;
}
#endif
