#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include "kmdw_console.h"
#include "kdrv_log.h"

#define BACKSP_KEY 0x08
#define RETURN_KEY 0x0D
#define DELETE_KEY 0x7F
#define BELL 0x07

bool UART0_Rx = false;
bool UART0_Tx = false;

static uint32_t scpu_debug_flags = 0;

static kdrv_uart_handle_t handle0;

static void UART0_callback(uint32_t event)
{
    if (event & ARM_USART_EVENT_RECEIVE_COMPLETE)
    {
        UART0_Rx = true;
    }

    if (event & ARM_USART_EVENT_RX_TIMEOUT)
    {
        UART0_Rx = true;
    }
}

void kmdw_console_init(kdrv_uart_dev_id_t uart_dev)
{
    kdrv_status_t sts = kdrv_uart_open(&handle0, uart_dev, UART_MODE_SYNC_RX | UART_MODE_SYNC_TX, UART0_callback);
    if (sts != KDRV_STATUS_OK)
    {
        dbg_msg("Open failed\n");
        return;
    }

    kdrv_uart_config_t cfg;
    cfg.baudrate = BAUD_115200; // BAUD_921600;
    cfg.data_bits = 8;
    cfg.frame_length = 0;
    cfg.stop_bits = 1;
    cfg.parity_mode = PARITY_NONE;
    cfg.fifo_en = false;

    sts = kdrv_uart_configure(handle0, UART_CTRL_CONFIG, (void *)&cfg);
    if (sts != KDRV_STATUS_OK)
    {
        dbg_msg("UART config failed\n");
        return;
    }
#ifdef USE_LOGGER    
    kdrv_logger_thread_create(handle0);
#endif
    UART0_Rx = false;
    UART0_Tx = false;
}

static char uart_buffer[256];
const char uart_error_msg[30] = "ERROR: UART Buffer Overrun.\r\n\0";
void kmdw_printf(const char *fmt, ...) /* variable arguments */
{
    va_list arg_ptr;

    va_start(arg_ptr, fmt);
    vsprintf(&uart_buffer[0], fmt, arg_ptr);
    va_end(arg_ptr);
    
    UART0_Tx = false;
    if (strlen(uart_buffer) > 256)
    {
        kdrv_uart_write(handle0, (uint8_t *)uart_error_msg, strlen(uart_error_msg));
        uart_buffer[255] = 0; // truncate the original output
        abort();
    }
#ifdef USE_LOGGER    
    kdrv_log_push(&uart_buffer[0]);
#else
    kdrv_uart_write(handle0, (uint8_t *)uart_buffer, strlen(uart_buffer));
#endif
}

void kmdw_level_printf(int level, const char *fmt, ...) /* variable arguments */
{
    uint32_t lvl = kmdw_console_get_log_level_scpu();
    lvl >>= 16;

    if ((level == LOG_PROFILE && level == lvl) || (lvl < LOG_PROFILE && level > 0 && level <= lvl))
    {
        va_list arg_ptr;

        va_start(arg_ptr, fmt);
        vsprintf(&uart_buffer[0], fmt, arg_ptr);
        va_end(arg_ptr);

        UART0_Tx = false;
        if (strlen(uart_buffer) > 256)
        {
            kdrv_uart_write(handle0, (uint8_t *)uart_error_msg, strlen(uart_error_msg));
            uart_buffer[255] = 0; // truncate the original output
        }
#ifdef USE_LOGGER    
        kdrv_log_push(&uart_buffer[0]);
#else
        kdrv_uart_write(handle0, (uint8_t *)uart_buffer, strlen(uart_buffer));
#endif

    }
}

char kmdw_console_getc(void)
{
    char c;
    UART0_Rx = false;
    kdrv_uart_read(handle0, (uint8_t *)&c, 1);
    return c;
}

void kmdw_console_putc(char Ch)
{
    char cc;

    if (Ch != '\0')
    {
        cc = Ch;
        kdrv_uart_write(handle0, (uint8_t *)&cc, 1);
    }

    if (Ch == '\n')
    {
        cc = '\r';
        kdrv_uart_write(handle0, (uint8_t *)&cc, 1);
    }
}

void kmdw_console_puts(char *str)
{
    char *cp;
    for (cp = str; *cp != 0; cp++)
        kmdw_console_putc(*cp);
}

int kmdw_console_echo_gets(char *buf, int len)
{
    char *cp;
    char data;
    uint32_t count;
    count = 0;
    cp = buf;

    do
    {
        kdrv_uart_get_char(handle0, &data);
        switch (data)
        {
        case RETURN_KEY:
            if (count < len)
            {
                *cp = '\0';
                kmdw_console_putc('\n');
            }
            break;
        case BACKSP_KEY:
        case DELETE_KEY:
            if ((count > 0) && (count < len))
            {
                count--;
                *(--cp) = '\0';
                kmdw_console_puts("\b \b");
            }
            break;
        default:
            if (data > 0x1F && data < 0x7F && count < len)
            {
                *cp = (char)data;
                cp++;
                count++;
                kmdw_console_putc(data);
            }
            break;
        }
    } while (data != RETURN_KEY);

    return (count);
}

__weak void kdrv_ncpu_set_scpu_debug_lvl(uint32_t lvl)
{
}

__weak void kdrv_ncpu_set_ncpu_debug_lvl(uint32_t lvl)
{
}

void kmdw_console_set_log_level_scpu(uint32_t level)
{
    scpu_debug_flags = (scpu_debug_flags & ~0x000F0000) | (((level) << 16) & 0x000F0000);
    kdrv_ncpu_set_scpu_debug_lvl(level);
}

uint32_t kmdw_console_get_log_level_scpu(void)
{
    return scpu_debug_flags;
}

void kmdw_console_set_log_level_ncpu(uint32_t level)
{
    kdrv_ncpu_set_ncpu_debug_lvl(level);
}
