#ifndef __MODEL_MGR_H__
#define __MODEL_MGR_H__

#include <stdint.h>
#include "cmsis_os2.h"
#include "ipc.h"
#include "base.h"
#include "model_type.h"

#define KMDW_MODEL_ALL_MODELS  -1

/* flash memory table
 *    BOOT related       :  192K(reserved)      0x0         - 0x2FFFF
 *    FLASH_ADDR_HEAD    :                      0x30000
 *    FID database       :  512K(reserved)      0x80000     - 0xFFFFF
 *    fw_info.bin        :    4K(reserved)      0x100000    - 0x100FFF  : real use: depend on model count
 *      - model count : 4 bytes
 *      - models      :sizeof(struct kdp_model_s)*count
 *      - ddr end addr:4 bytes
 *
 *    bin files for models: (4K aligned)        0x101000     - ??       : real use: depend model size
 *      - cmd.bin + weight.bin + setup.bin
 *    FLASH_ADDR_END                            0x101000 + model size   -> + 4K alignment
 */

/* size of fw_info.bin & all_models.bin */
#define MODEL_MGR_SIZE_MODEL_INFO  224      // (232 - 4 -4 )   todo: must specify the size
#define MODEL_MGR_SIZE_MODEL_POOL  3572576  // todo: musth specify the size

#define MODEL_MGR_FLASH_ADDR_HEAD               0x00300000  // expand DB from 512k to 2560k
#define MODEL_MGR_FLASH_ADDR_MODEL_COUNT        (MODEL_MGR_FLASH_ADDR_HEAD)
#define MODEL_MGR_FLASH_ADDR_MODEL_INFO         (MODEL_MGR_FLASH_ADDR_MODEL_COUNT + 4)
#define MODEL_MGR_FLASH_ADDR_MODEL_DDR_END_ADDR (MODEL_MGR_FLASH_ADDR_MODEL_INFO  + MODEL_MGR_SIZE_MODEL_INFO)
#define MODEL_MGR_FLASH_ADDR_MODEL_POOL         (MODEL_MGR_FLASH_ADDR_HEAD + 0x1000)

/* These crc values will ONLY change if model is changed in FLASH */
#define MODEL_INFO_CRC16_VALUE  0x5341
#define MODEL_INFO_SUM32_VALUE  0x00eee074

#ifndef FLASH_MODEL_HAS_CRC_CHECK
#define MODEL0_FD_CRC16_VALUE   0x939f   /* original: 0x4318 */
#define MODEL1_LM_CRC16_VALUE   0x1cd0   /* original: 0xfce0 */
#define MODEL3_FR_CRC16_VALUE   0x5df0   /* original: 0xb359 */
#define MODEL0_FD_SUM32_VALUE   0xf9e571d8
#define MODEL1_LM_SUM32_VALUE   0x3f96c4fc
#define MODEL3_FR_SUM32_VALUE   0xbc998f45
#endif

enum kmdw_model_rc {
    // 0 - 9 is reserved for ncpu return
    // defined in ipc.h
    // IMAGE_STATE_INACTIVE == 0
    // IMAGE_STATE_ACTIVE == 1
    // IMAGE_STATE_NPU_DONE == 2
    // IMAGE_STATE_DONE == 3
    KMDW_MODEL_RUN_RC_ABORT = 10,
    KMDW_MODEL_RUN_RC_ERROR = 11,
    KMDW_MODEL_RUN_RC_END
};

struct img_run_time_s {
    uint32_t    round_trip_time;
    uint32_t    pre_proc_time;
    uint32_t    npu_proc_time;
    uint32_t    post_proc_time;
};

/* ############################
 * ##    Public Functions    ##
 * ############################ */

/**
 * @brief Init model functionality
 */
void kmdw_model_init(void);

/**
 * @brief A wrapper of load_model
 * @param model_index_p: model info index,
 *                    0-n: info_index of model to load
 *                    -1 means to load all models
 * @return always 0
 */
int32_t kmdw_model_load_model(int8_t model_info_index_p);

/**
 * @brief A wrapper of load_model_info
 * @param [in] is_model_from_ddr: if model is from ddr/host command
 */
void kmdw_model_reload_model_info(bool from_ddr);

/**
 * @brief Refresh all models
 */
void kmdw_model_refresh_models(void);

/**
 * @brief Output model_info of specified index
 * @param[in] idx_p the index of programmed models
 * @return model_info defined in ipc.h
 */
struct kdp_model_s* kmdw_model_get_model_info(int idx_p);

/**
 * @brief Specify output address for model run in ncpu/npu
          !!! must be called after kmdw_model_config_model()
 * @return always 0
 */
int32_t kmdw_model_config_result(osEventFlagsId_t result_evt, uint32_t result_evt_flag);

/**
 * @brief Config model image
 * @param img_cfg image config
 * @param crop_box image crop config
 * @param pad_values image padding config
 * @param ext_param extra param
 */
void kmdw_model_config_img(struct kdp_img_cfg *img_cfg, struct kdp_crop_box_s *crop_box,
                           struct kdp_pad_value_s *pad_values, void *ext_param);

/**
 * @brief Get raw image config
 * @param idx image index
 * @return raw image config
 */
struct kdp_img_raw_s* kmdw_model_get_raw_img(int idx);

/**
 * @brief Run model
 * @param tag model tag
 * @param output model output
 * @param model_type model type
 * @param dme DME mode
 * @return kmdw_model_rc
 */
int kmdw_model_run(const char *tag, void *output, uint32_t model_type, bool dme);

/**
 * @brief Abort model execution
 */
void kmdw_model_abort(void);

/**
 * @brief Get round-trip/pre/npu/post times
 * @param img_idx raw image index
 * @param run_time a pointer of struct img_run_time_s to receive time
 * @return void
 */
void kmdw_model_get_run_time(int img_idx, struct img_run_time_s *run_time);

#if DEBUG

/**
 * @brief Dump model debug info
 */
void kmdw_model_dump_model_info(void);

#endif

#endif
