/**
 * Copyright (C) Kneron Inc. All right reserved
 *
 * @file kapp.h
 * @brief Kneron Application Shared Data
 */

#ifndef __KAPP_H_
#define __KAPP_H_

#include <stdint.h>
#include "model_type.h"
#include "ipc.h"
#include "kmdw_model.h"

/* #define DUMP_IMG */

enum kapp_status_code_e {
    KAPP_OK = 0,
    KAPP_UNKNOWN_ERR = 1
};

enum kapp_model_status_code_e {
    MODEL_APP_OK = 0,
    MODEL_APP_UNKNOWN_ERR,
    MODEL_APP_ABORT,
    MODEL_APP_FDR_WRONG_USAGE = 0x100, // MODEL_APP_FID_CODES
    MODEL_APP_FDR_NO_FACE,
    MODEL_APP_FDR_BAD_POSE,
    MODEL_APP_FDR_FR_FAIL,
    MODEL_APP_LW3D_WRONG_USAGE = 0x300, //MODEL_APP_LW3D_CODES
    MODEL_APP_LW3D_NO_ALIVE,
};

/* no need to align with cam source index */
#define KAPP_IMG_SRC_RGB  0
#define KAPP_IMG_SRC_NIR  1
#define KAPP_IMG_SRC_LV   2
#define KAPP_IMG_SRC_MAX  3

#define KAPP_IMAGE_FORMAT_RGBA8888   0x80000000   /**< image format: RGBA8888 */
#define KAPP_IMAGE_FORMAT_RAW8       0x80000020   /**< image format: RAW8 */
/* The data byte sequence in DDR memory is [lowest byte]Y0CbY1CrY2CbY3Cr...[highest byte] */
#define KAPP_IMAGE_FORMAT_YCBCR422   0x80000037   /**< image format: YCBCR422 */
#define KAPP_IMAGE_FORMAT_RGB565     0x80000060   /**< image format: RGB565 */
#define KAPP_IMAGE_FORMAT_LV_NCPU    0x80040020

/* kai2do: can't merge w/ kdp_img_cfg now since changing
 *         kdp_dme_cfg::img_conf needs to change client code as well */
/**
 * @brief  KDP image configuration structure
 */
struct kdp_img_conf_s {
    int32_t     image_col;        /// column size
    int32_t     image_row;        /// row size
    int32_t     image_ch;         /// channel size
    uint32_t    image_format;     /// image format. KAPP_IMAGE_FORMAT_XXXX
    uint32_t    image_mem_addr;   /// address of image data
};

struct kapp_model_ext_params { //mapping "strcut parameter_s"
    uint32_t params[10];
};

struct kdp_ext_dual_landmarks_s { // larry lai, for liveness
    uint32_t    params[20];
};

/* #####################
 * ##    functions    ##
 * ##################### */

/**
 * @brief generate image config
 * @param img_buf_idx active buffer index
 * @return image config
 */
struct kdp_img_cfg kapp_gen_img_cfg(int32_t col, int32_t row, int32_t ch,
                                    uint32_t fmt, uint32_t addr, uint32_t img_buf_idx);

/**
 * @brief get image config
 * @param img_src image source index
 * @return image config
 */
struct kdp_img_cfg* kapp_get_img_cfg(int img_src);

/**
 * @brief To config image setting
 *
 * @param [in] src_index src img index (no need to align with cam src)
 * @param [in] col column size
 * @param [in] row row size
 * @param [in] ch channel size
 * @param [in] fmt propietory image format
 * @param [in] mem_addr image DDR address
 * @param [in] img_buf_idx raw image buffer index
 */
void kapp_config_image(int src_img, int32_t col, int32_t row, int32_t ch,
                       uint32_t fmt, uint32_t mem_addr, int img_buf_idx);

/**
 * @brief to update image memory address
 *
 * !!! MUST be called after kapp_config_image()
 * @param [in] img source rgb/nir
 * @param [in] image_memory_address new image memory address
 * @return n/a
 */
void kapp_set_img_mem_addr(int img_src_idx, int32_t img_mem_addr);

/**
 * @brief output image memory address of specified source
 * @param [in] camera image source (RGB or NIR)
 * @return image memory address being used currently
 */
uint32_t kapp_get_img_mem_addr(int img_src_idx);

/**
 * @brief to get channel size by image format
 *
 * @param [in] image format
 * @return channel size
 */
uint32_t kapp_get_chnl_size(uint32_t format);

/**
 * @brief to get pixel size by image format
 *
 * @param [in] image format
 * @return pixel size
 */
uint32_t kapp_get_pixel_size(uint32_t image_format);

/**
 * @brief to switch DME mode
 *
 * @param [in] is_dme
 * @return n/a
 */
void kapp_set_dme_mode(bool is_dme);

#endif
