/*
 * Kneron Host Communication driver
 *
 * Copyright (C) 2018 Kneron, Inc. All rights reserved.
 *
 */
#include <stdio.h>
#include <string.h>
#include "base.h"
#include "def_share.h"
#include "com.h"
#include "com_err.h"
#include "msg.h"
#include "kmdw_console.h"
#include "kcomm.h"
#include "kapp.h"
#include "kapp_fid.h"
#include "kapp_lw3d.h"
#include "kapp_db.h"
#include "kmdw_memory.h"

#define DEF_LOG_CATEG  "kcomm"

enum FID_DB_OP_MODE {
    DB_IMPORT = 1,
    DB_EXPORT = 2,
    DB_QUERY = 3,
    DB_ADD_FM = 4,
    DB_DEL_FM = 5,
    DB_OP_MAX = 6
};

static u16 SFID_user;   // SFID user id
static u16 SFID_index;  // SFID image index
static kapp_image_buffer_desc_t SFID_buf_desc;
static float RGB_threshold = FID_THRESHOLD;
static float NIR_threshold = FID_THRESHOLD;

static struct facedet_result_s  *rgb_fd_out = 0;
static struct landmark_result_s *rgb_lm_out = 0;
static struct fr_result_s       *rgb_fr_out = 0;

static struct facedet_result_s  *nir_fd_out = 0;
static struct landmark_result_s *nir_lm_out = 0;
static struct fr_result_s       *nir_fr_out = 0;

static kapp_fid_inout_t         *lw3d_out = 0;

extern int kdp_memxfer_ddr_to_flash(u32 dst, u32 src, size_t bytes);

/* ###########################
 * ##    local functions    ##
 * ########################### */

static int config_img(int src_img_idx, struct kdp_img_cfg *img_cfg, kapp_image_buffer_desc_t *buf_desc)
{
    int fmt = img_cfg->image_format;
    int pixel_size = kapp_get_pixel_size(fmt);
    int ch = kapp_get_chnl_size(fmt);
    int img_buf_idx;

    if (pixel_size == 0) {
        return -1;
    }

    img_cfg->image_mem_len = img_cfg->image_col * img_cfg->image_row * pixel_size;

    buf_desc->buf_count = KAPP_IMG_SRC_MAX;
    buf_desc->alignment_in_bit  = 32;

    if (src_img_idx == KAPP_IMG_SRC_RGB) {
        buf_desc->rgb_buf_base_addr = img_cfg->image_mem_addr;
        buf_desc->rgb_img_size = img_cfg->image_mem_len;
        img_buf_idx = 0;
    } else if (src_img_idx == KAPP_IMG_SRC_NIR) {
        buf_desc->nir_buf_base_addr = img_cfg->image_mem_addr;
        buf_desc->nir_img_size = img_cfg->image_mem_len;
        img_buf_idx = 1;
    }

    kapp_config_image(src_img_idx, img_cfg->image_col, img_cfg->image_row, ch,
                         fmt, img_cfg->image_mem_addr, img_buf_idx);

    return 0;
}

static int fid_ret_to_err(int ret)
{
    switch (ret) {
    case KAPP_DB_NO_MATCH:
        return MSG_APP_DB_NO_MATCH;
    case KAPP_DB_ALREADY_SAVED:
        return MSG_APP_UID_EXIST;
    case KAPP_DB_REG_FIRST:
        return MSG_APP_IDX_FIRST;
    case KAPP_DB_USER_NOT_REG:
        return MSG_APP_IDX_MISSING;
    case KAPP_DB_NO_SPACE:
        return MSG_APP_DB_FAIL;
    case KAPP_FDR_WRONG_USAGE:
        return MSG_APP_BAD_INDEX;
    case KAPP_DB_DEL_NOT_VALID:
        return MSG_APP_DB_NO_USER;
    default:
        return MSG_APP_BAD_IMAGE;
    }
}

static bool valid_app_id(int id)
{
    return ((id > APP_UNKNOWN) && (id < APP_ID_MAX));
}

static u16* set_img_cfg(struct kdp_img_cfg *cfg, u16 *data)
{
    u16 col = *data++;
    u16 row = *data++;
    u32 *pword = (u32*)data;
    u32 fmt = *pword++;

    cfg->image_col = col;
    cfg->image_row = row;
    cfg->image_format = fmt;

    dlog("img col [%u], row [%u], fmt [0x%x]", col, row, fmt);

    return (u16*) pword; // final pointer pos
}

static void output_results(u16 *output_buf, void *out_result, u32 count)
{
    u32 *result = out_result;
    while (count) {
        *output_buf++ = (u16) *result++;
        count--;
    }
}

/* This function checks the output data and if valid, will append them to the message
** input:  output flags to check, pointer to start of output data in message buffer
** output: number of bytes added to the message buffer
*/
static u32 fdlmfrlv_outputs(u32 *check_flag, u16 *pfdfr_out, u32 lv_flag)
{
    u32 out_size = 0;
    u32 out_flag = *check_flag;

    if (*(u32*)rgb_fd_out) { // FD valid?
        if (out_flag & 1) {  // do we export FD?
            output_results(pfdfr_out, &rgb_fd_out->xywh[0], (FD_RES_LENGTH / 2)); // u16 output, 4 members
            out_size += FD_RES_LENGTH;
            pfdfr_out += (FD_RES_LENGTH / 2);
        }
    } else {
        out_flag &= 0x0fff0;  // no FD, then no LM, FR, LV either
    }

    if (out_flag & 2) {  // export LM?
        if (*(u32*)rgb_fd_out) { // if FD is good, assume LM is also good
            output_results(pfdfr_out, &rgb_lm_out->marks[0].x, (LM_RES_LENGTH / 2)); // 10 members
            out_size += LM_RES_LENGTH;
            pfdfr_out += (LM_RES_LENGTH / 2);
        }
    }
    if (out_flag & 4) {  // export FR?
        if (*(u32*)rgb_fr_out) { // FR valid?
            memcpy((u8*)pfdfr_out, (u8*)rgb_fr_out->feature_map, FR_RES_LENGTH);
            out_size += FR_RES_LENGTH;
            pfdfr_out += (FR_RES_LENGTH / 2);
        }
        else
            out_flag &= 0x0fffb;
    }
    if (out_flag & 8) {  // export LV?
        if (lv_flag & 0x80) { // LV valid?
            *(u32*)pfdfr_out = lv_flag & 1;  // store the results
            out_size += LV_RES_LENGTH;
            pfdfr_out += (LV_RES_LENGTH / 2);
        }
        else
            out_flag &= 0x0fff7;
    }
    if (*(u32*)nir_fd_out) { // nir FD valid?
       if (out_flag & 0x10) {  // do we export nir FD?
            output_results(pfdfr_out, &nir_fd_out->xywh[0], (FD_RES_LENGTH / 2)); // u16 output, 4 members
            out_size += FD_RES_LENGTH;
            pfdfr_out += (FD_RES_LENGTH / 2);
        }
    }
    else
        out_flag &= 0x0fe0f;  // no FD, then no LM, FR, LV, Final LV either
    if (out_flag & 0x20) {  // export nir LM?
        if (*(u32*)nir_fd_out) { // if FD is good, assume LM is also good
            output_results(pfdfr_out, &nir_lm_out->marks[0].x, (LM_RES_LENGTH / 2)); // 10 members
            out_size += LM_RES_LENGTH;
            pfdfr_out += (LM_RES_LENGTH / 2);
        }
    }
    if (out_flag & 0x40) {  // export FR?
        if (*(u32*)nir_fr_out) { // FR valid?
            memcpy((u8*)pfdfr_out, (u8*)nir_fr_out->feature_map, FR_RES_LENGTH);
            out_size += FR_RES_LENGTH;
            pfdfr_out += (FR_RES_LENGTH / 2);
        }
        else
            out_flag &= 0x0ffbf;
    }
     if (out_flag & 0x100) {  // export final LV?
        if (lv_flag & 0x80) { // NIR LV result valid?
            *(u32*)pfdfr_out = lv_flag & 1;  // store the results
            out_size += LV_RES_LENGTH;
            pfdfr_out += (LV_RES_LENGTH / 2);
        }
        else
            out_flag &= 0x0feff;
    }
    *check_flag = out_flag;
    return out_size;
}

static void get_lw3d_io_memory(void)
{
    if(lw3d_out == 0)
        lw3d_out = (kapp_fid_inout_t*)kmdw_ddr_reserve(sizeof(kapp_fid_inout_t));
    if(rgb_fd_out == 0)
        rgb_fd_out = (struct facedet_result_s*)kmdw_ddr_reserve(sizeof(struct facedet_result_s));
    if(rgb_lm_out == 0)
        rgb_lm_out = (struct landmark_result_s*)kmdw_ddr_reserve(sizeof(struct landmark_result_s));
    if(rgb_fr_out == 0)
        rgb_fr_out = (struct fr_result_s*)kmdw_ddr_reserve(sizeof(struct fr_result_s));
    if(nir_fd_out == 0)
        nir_fd_out = (struct facedet_result_s*)kmdw_ddr_reserve(sizeof(struct facedet_result_s));
    if(nir_lm_out == 0)
        nir_lm_out = (struct landmark_result_s*)kmdw_ddr_reserve(sizeof(struct landmark_result_s));
    if(nir_fr_out == 0)
        nir_fr_out = (struct fr_result_s*)kmdw_ddr_reserve(sizeof(struct fr_result_s));

}

/* ########################
 * ##    cmd handlers    ##
 * ######################## */

static void handle_fid_start(RspPram *rsp, MsgHdr *hdr,
                             float rgb_thresh_val, float nir_thresh_val, u8 *data)
{
    int ret = 0;
    int app_type = (hdr->cmd == CMD_SFID_LW3D_START) ? APP_LW3D : APP_SFID;
    float rgb_thresh = rgb_thresh_val ? rgb_thresh_val : FID_THRESHOLD;
    float nir_thresh = nir_thresh_val ? nir_thresh_val : FID_THRESHOLD;

    dlog("app_type [%d], rgb_thresh [%.3f], nir_thresh [%.3f], data [0x%x]",
          app_type, rgb_thresh, nir_thresh, data);

    kapp_set_dme_mode(false);

    if (SFID_user) {
        ret = kapp_db_abort_reg();
        if (ret != KAPP_OK) {
            err_msg("[kcomm] Start Abort Old Session FAILED.\r\n");
        }
    }

    info_msg("[kcomm] Start FD/FR/DB.\r\n");
    SFID_user = 0; // switch to NORMAL FD/FR/DB MODE
    SFID_index = 0;

    // config image settings
    if (hdr->msg_len > 8) {
        u16 *prd = set_img_cfg(&rgb_img_config, (u16*)data);
        if (app_type == APP_LW3D) {
            set_img_cfg(&nir_img_config, prd);
        }
    }

    if (app_type == APP_LW3D) {
        ret = (0 == config_img(KAPP_IMG_SRC_RGB, &rgb_img_config, &SFID_buf_desc) &&
               0 == config_img(KAPP_IMG_SRC_NIR, &nir_img_config, &SFID_buf_desc)) ? 0 : -1;
    } else {
        ret = config_img(KAPP_IMG_SRC_RGB, &rgb_img_config, &SFID_buf_desc);
    }

    if (ret != 0) {
        rsp->error = MSG_CONFIG_ERROR;
        rsp->param2 = 0; // image size set to 0
    } else {
        dbg_msg("[kcomm] FID START, buf count [%u], alignment [%u], "
                "rgb addr [0x%x], size [%u], nir addr [0x%x], size [%u]\r\n",
                SFID_buf_desc.buf_count, SFID_buf_desc.alignment_in_bit,
                SFID_buf_desc.rgb_buf_base_addr, SFID_buf_desc.rgb_img_size,
                SFID_buf_desc.nir_buf_base_addr, SFID_buf_desc.nir_img_size);

        APP_id = app_type;
        RGB_threshold = rgb_thresh;
        NIR_threshold = nir_thresh;

        // start continuous inference & DB match
        rsp->error = NO_ERROR;
        rsp->param2 = SFID_buf_desc.rgb_img_size;
    }

    if (app_type == APP_LW3D) {
        u32 *pword = (u32*)data; // use rcv buffer to format responses
        if (ret) {
            *pword = 0;
        } else {
            *pword = SFID_buf_desc.nir_img_size;
        }
        kcomm_send_rsp(CMD_SFID_LW3D_START, data, 4);
    } else {
        kcomm_send_rsp(CMD_SFID_START, NULL, 0);
    }
}

static void handle_fid_new_user(RspPram *rsp, u32 uid, u32 idx)
{
    dlog("uid [%u], idx [%u]", uid, idx);

    if ((uid == 0) || !valid_app_id(APP_id)) {
        rsp->error = (uid == 0) ? MSG_SFID_USR_ZERO : MSG_CONFIG_ERROR;
        rsp->param2 = 0;
        kcomm_send_rsp(CMD_SFID_NEW_USER, NULL, 0);
        return;
    }

    if (SFID_user == 0) {  // switch to ADD USER MODE
        SFID_user = uid;
        SFID_index = idx;
    } else {
        if (SFID_user == uid) {
            SFID_index = idx;
        } else {
            if (idx == 1) {
                // abort any previous NEW USER session first
                int ret = kapp_db_abort_reg();
                if (ret == KAPP_OK) {
                    SFID_user = uid;
                    SFID_index = 1;
                } else { // shouldn't happen
                    err_msg("[kcomm] Add user Abort Old Session %d Failed -> %d\r\n", uid, ret);
                    // don't know how to handle this, leave to image cmd
                    SFID_user = uid; // force the case anyway
                }
            } else { // we don't send error code out anymore
                SFID_user = uid;  // let image cmd deal with the problem
                SFID_index = idx;
            }
        }
    }

    info_msg("[kcomm] Add New User %d Indx -> %d\r\n", SFID_user, SFID_index);

    rsp->error = NO_ERROR;
    rsp->param2 = 0;
    kcomm_send_rsp(CMD_SFID_NEW_USER, NULL, -1);  // use zero payload response
}

static void handle_send_img(RspPram *rsp, u32 img_size, u32 out_flag)
{
    dlog("out_flag [0x%x], img_size [%u]", out_flag, img_size);

    u32 out_size, lv_flag = 0;  // lv_flag (byte) bit 7-> result_valid, bit 0-> result
    u16 *pfdfr_out;

    out_flag &= 0x0F;  // ---- ---Z vrmd VRMD -> final LV, rsv.fr.lm.fd, rgbLV.FR.LM.FD
    pfdfr_out = (u16*)rsp->data;  // point to output data field
    *pfdfr_out++ = 0;  // initialize the response msg data fields
    *pfdfr_out++ = 0;  // now pointing to the optional results field

    if (APP_id != APP_SFID) {
        rsp->error = MSG_CONFIG_ERROR;
        rsp->param2 = CMD_SFID_SEND_IMAGE;
        kcomm_send_rsp(CMD_ACK_NACK, NULL, 0);
        return;
    }

    get_lw3d_io_memory();
    struct facedet_result_s *fd_out = rgb_fd_out;
    struct landmark_result_s *lm_out = rgb_lm_out;
    struct fr_result_s *fr_out = rgb_fr_out;
    kapp_fid_inout_t *fdr_out = lw3d_out;

    if (SFID_index == 0) {
        fdr_out->op = KAPP_FID_OP_INFERENCE;
        fdr_out->face_id = 0; //inference mode
    } else {
        fdr_out->op = KAPP_FID_OP_REGISTER;
        fdr_out->face_id = SFID_index; //reg mode, fm index
    }
    fdr_out->user_id = SFID_user;
    fdr_out->thresh_fid[KAPP_IMG_SRC_RGB] = RGB_threshold; //threshold value for FID db compare

    // get image
    int ret = kcomm_read(SFID_buf_desc.rgb_buf_base_addr, img_size);  // specify how many bytes
    if (ret == img_size) { // got image
        // input range check first
        if ((SFID_user != 0) && (SFID_index != 0)) {
            if ((SFID_user > MAX_USER) || (SFID_index > MAX_FID)) {
                // bad user id or image index
                rsp->error = MSG_APP_BAD_INDEX;
                rsp->param2 = 0;
                kcomm_send_rsp(CMD_SFID_SEND_IMAGE, rsp->data, 4);
                return;
            }
        }

        // now send command to AppMgr
        *(u32*)fd_out = 0;  // initialize the results, FD and FR only
        *(u32*)fr_out = 0;  // fdr() does not perform LV right now
        ret = kapp_fdr(KAPP_IMG_SRC_RGB, fdr_out, fd_out, lm_out, fr_out);
        if (ret == KAPP_OK) {
            u16* mode = (u16*) &rsp->param2;
            rsp->error = NO_ERROR; // got user_id or feature map stored
            // we have u16 mode and u16 user_id/fm_indx in a u32 field (Little Endian)
            if (SFID_user) {
                *mode++ = 1;
                *mode = SFID_index;
            } else {
                *mode++ = 0;
                *mode = fdr_out->user_id;
            }
        } else { // error conditions
            rsp->error = fid_ret_to_err(ret);
        }
        out_size = fdlmfrlv_outputs(&out_flag, pfdfr_out, lv_flag);
        pfdfr_out = (u16*)&rsp->data;
        *pfdfr_out = out_flag;  // update output flag
        kcomm_send_rsp(CMD_SFID_SEND_IMAGE, rsp->data, 4 + out_size);
    } else { // no image
        rsp->error = MSG_APP_BAD_IMAGE;
        rsp->param2 = 0;
        if (ret < 0) {
            kcomm_send_no_rsp();  // if tx/rx timed out, abort the msg reply
        } else {
            kcomm_send_rsp(CMD_SFID_SEND_IMAGE, rsp->data, 4);
        }
    }
}

static void handle_send_lw3d_img(RspPram *rsp, u32 rgb_size, u32 nir_size, u16 *out_flag_val)
{
    dlog("rgb_size [%u], nir_size [%u], out_flag [%u]", rgb_size, nir_size, *out_flag_val);

    u32 out_size, lv_flag = 0;  // lv_flag (byte) bit 7-> result_valid, bit 0-> result
    u32 out_flag = *out_flag_val & 0x177; // ---- ---Z vrmd VRMD -> final LV, rsv.fr.lm.fd, rgbLV.FR.LM.FD
    u16 *pfdfr_out = (u16*)rsp->data;  // point to output data field
    *pfdfr_out++ = 0;  // initialize the response msg data fields
    *pfdfr_out++ = 0;  // now pointing to the optional results field

    if (APP_id != APP_LW3D) {
        rsp->error = MSG_CONFIG_ERROR;
        rsp->param2 = CMD_SFID_LW3D_IMAGE;
        kcomm_send_rsp(CMD_ACK_NACK, NULL, 0);
        return;
    }

    get_lw3d_io_memory();

    if (SFID_index == 0) {
        lw3d_out->op = KAPP_FID_OP_INFERENCE;
        lw3d_out->face_id = 0; //inference mode
    } else {
        lw3d_out->op = KAPP_FID_OP_REGISTER;
        lw3d_out->face_id = SFID_index; //reg mode, fm index
    }

    lw3d_out->user_id = SFID_user;
    lw3d_out->thresh_fid[KAPP_IMG_SRC_RGB] = RGB_threshold;
    lw3d_out->thresh_fid[KAPP_IMG_SRC_NIR] = NIR_threshold;

    dbg_msg("[kcomm] wait for lw3d images\r\n");

    int ret = kcomm_read(SFID_buf_desc.rgb_buf_base_addr, rgb_size);
    if (ret == SFID_buf_desc.rgb_img_size) {
        ret = kcomm_read(SFID_buf_desc.nir_buf_base_addr, nir_size);
    } else {
        // send NACK to abort nir image downloading
        rsp->error = MSG_APP_BAD_IMAGE; // image not expected size
        rsp->param2 = CMD_SFID_LW3D_IMAGE; // add source id
        kcomm_send_rsp(CMD_ACK_NACK, NULL, 0);
        return;
    }

    if (ret != SFID_buf_desc.nir_img_size) { // got nir image as well
        rsp->error = MSG_APP_BAD_IMAGE;
        rsp->param2 = 0;
        if (ret < 0) {
            kcomm_send_no_rsp();  // if tx/rx timed out, abort the msg reply
        } else {
            kcomm_send_rsp(CMD_SFID_LW3D_IMAGE, rsp->data, 4);
        }
        return;
    }

    if ((SFID_user != 0) && (SFID_index != 0)) {
        if ((SFID_user > MAX_USER) || (SFID_index > MAX_FID)) {
            rsp->error = MSG_APP_BAD_INDEX;
            rsp->param2 = 0;
            kcomm_send_rsp(CMD_SFID_LW3D_IMAGE, rsp->data, 4);
            return;
        }
    }

    // initialize the results, FD and FR only
    *(u32*)rgb_fd_out = 0;
    *(u32*)rgb_fr_out = 0;
    *(u32*)nir_fd_out = 0;
    *(u32*)nir_fr_out = 0;

    bool liveness = false;
    ret = kapp_lw3d(lw3d_out, rgb_fd_out, rgb_lm_out, rgb_fr_out,
                    nir_fd_out, nir_lm_out, nir_fr_out, &liveness);
    u16* mode = (u16*) &rsp->param2; // point to the mode field in the rsp msg

    if (ret == KAPP_OK) {
        // got user_id or new user feature map stored
        rsp->error = NO_ERROR;

        // we have u16 mode and u16 user_id/fm_indx in a u32 field (Little Endian)
        if (SFID_user) {
            *mode++ = 1;
            *mode = SFID_index;
        } else {
            *mode++ = 0;
            *mode = lw3d_out->user_id;
        }
    } else { // error conditions
        trace_msg("[host_com] lw3d ret = 0x%x\r\n", ret);
        rsp->error = fid_ret_to_err(ret);
        rsp->param2 = 0;  // the mode fields has no meaning
    }

    lv_flag = liveness ? 0x81 : 0x80;
    out_size = fdlmfrlv_outputs(&out_flag, pfdfr_out, lv_flag);
    pfdfr_out = (u16*)&rsp->data;
    *pfdfr_out = out_flag;  // update output flag
    kcomm_send_rsp(CMD_SFID_LW3D_IMAGE, rsp->data, 4 + out_size);
}

static void handle_sfid_add_db(RspPram *rsp)
{
    dlog("");

    if (!valid_app_id(APP_id)) {
        rsp->error = MSG_CONFIG_ERROR;  // not in SFID mode
        rsp->param2 = 0;
        kcomm_send_rsp(CMD_SFID_ADD_DB, NULL, 0);
        return;
    }

    kapp_db_user_data_t data;
    data.user_id_in = SFID_user;

    int ret = kapp_db_add(&data);
    if (ret == KAPP_OK) {  // success
        rsp->error = NO_ERROR;
    } else { // error conditions
        switch (ret) {
        case KAPP_DB_ALREADY_SAVED:
            rsp->error = MSG_APP_UID_EXIST;
            break;
        case KAPP_DB_REG_FIRST:
            rsp->error = MSG_APP_IDX_FIRST;
            break;
        case KAPP_DB_USER_NOT_REG:
            rsp->error = MSG_APP_IDX_MISSING;
            break;
        default:
            rsp->error = MSG_APP_DB_FAIL;
            break;
        }
    }
    rsp->param2 = 0;
    kcomm_send_rsp(CMD_SFID_ADD_DB, NULL, 0);
}

static void handle_sfid_del_db(RspPram *rsp, u32 uid)
{
    dlog("uid [%u]", uid);

    kapp_db_user_data_t data;

    if (!valid_app_id(APP_id)) {
        rsp->error = MSG_CONFIG_ERROR;  // not in SFID mode
        rsp->param2 = 0;
        kcomm_send_rsp(CMD_SFID_DELETE_DB, NULL, 0);
        return;
    }

    info_msg("[kcomm] DELETE USER %d\n", uid);

    data.user_id_in = uid;
    data.del_all = (uid == ALL_USER_UID) ? TRUE : FALSE;

    int ret = kapp_db_delete(&data);
    if (ret == KAPP_OK) {
        rsp->error = NO_ERROR;
    } else {
        if (ret == KAPP_DB_DEL_FAIL) {
            rsp->error = MSG_APP_DB_NO_USER;
        } else {
            rsp->error = MSG_APP_DB_FAIL;
        }
    }
    rsp->param2 = 0;
    kcomm_send_rsp(CMD_SFID_DELETE_DB, NULL, 0);
}

static void handle_sfid_edit_db(RspPram *rsp, u32 mode, u8 *param)
{
    dlog("mode [%u], param [0x%x]", mode, param);

    u32 *db_ddr_addr, *db_flash_addr;
    u32 db_size = kapp_db_query(&db_ddr_addr, &db_flash_addr);  // get DB size & location
    bool app_id_ok = valid_app_id(APP_id);
    u16 *puid_out = (u16*)rsp->data;  // point to output data field
    int i = 0;
    int ret = 0;

    if (mode == DB_IMPORT) {
        if (!app_id_ok) {
            rsp->error = MSG_CONFIG_ERROR;  // image processing not set up yet
            rsp->param2 = CMD_SFID_EDIT_DB;  // add source id
            kcomm_send_rsp(CMD_ACK_NACK, NULL, 0);
            return;
        }

        if (db_size == *(u32*)param) {
            ret = kcomm_read((u32)db_ddr_addr, db_size);  // get DB binary
            if (ret < db_size) {
                kcomm_send_no_rsp();  // if tx/rx timed out, abort the msg processing
                kapp_db_init((u32)db_flash_addr); // reload database from flash, addr needed if db was flushed
                return;
            }
            rsp->param2 = 0;
            if (ret == db_size) {
                kdp_memxfer_ddr_to_flash((u32)db_flash_addr, (u32) db_ddr_addr, db_size);  // update db in flash
                rsp->error = NO_ERROR;
            } else {
                rsp->error = MSG_DATA_ERROR;
                kapp_db_init((u32)db_flash_addr);  // need to reload database
            }
        } else {
            rsp->error = MSG_DB_BAD_PARAM;
            rsp->param2 = CMD_SFID_EDIT_DB;  // add source id
            kcomm_send_rsp(CMD_ACK_NACK, NULL, 0);
            return;
        }
    } else if ((mode < DB_OP_MAX) && !app_id_ok) {  // for sub command 2-5
        rsp->error = MSG_CONFIG_ERROR;  // image processing not set up yet
        rsp->param2 = 0;
    } else if (mode == DB_EXPORT) { // export DB
        rsp->error = 0;
        rsp->param2 = db_size;  // set size of DB for export
        kcomm_send_rsp(CMD_SFID_EDIT_DB, NULL, 0);  // send reply now
        kcomm_write((u8*)db_ddr_addr, db_size, WMODE_ALT);  // use alternate data mode
        return;
    } else if (mode == DB_QUERY) { // send DB user list
        rsp->error = 0;
        i = 0;
        for (ret = 0; ret < MAX_USER; ret++) { // get DB user info
            if (kapp_db_get_user_id_slot_type(ret) == TYPE_VALID) {
                *puid_out = kapp_db_get_user_id(ret);
                puid_out++;
                i++;
            }
        }
        rsp->param2 = i;  // user id count
        kcomm_send_rsp(CMD_SFID_EDIT_DB, rsp->data, i*2);  // send reply now
        return;
    } else if (mode == DB_ADD_FM) { // add single FM in DB
        i = *(u32*)param;  // get user id of request
        int uidx = kapp_db_get_user_slot(i);  // get DB slot
        if (uidx < 0) {  // user id not in DB
            uidx = kapp_db_get_available_ID_slot();
            if (uidx >= MAX_USER) {
                rsp->error = MSG_DB_ADD_FM_FAIL;
                rsp->param2 = MSG_SFID_NO_MEM;
                kcomm_send_rsp(CMD_SFID_EDIT_DB, NULL, 0);
                return;
            }
        }
        ret = kapp_db_add_FM(uidx, i, (u8*)(&param[4]));
        rsp->param2 = 0;
        if (ret > 0) { // add FM successful
            rsp->error = 0;
            rsp->param2 = ret;
        } else {
            rsp->error = MSG_DB_ADD_FM_FAIL;  // DB error occurred
        }
        if (ret == -1) {
            rsp->param2 = MSG_APP_UID_EXIST;  // user id mismatch
        } else if (ret == -2) {
            rsp->param2 = MSG_APP_BAD_INDEX;
        }
    } else if (mode == DB_DEL_FM) { // delete single FM in DB
        i = *(u16*)param;  // get user id of request
        int uidx = kapp_db_get_user_slot(i);  // get DB slot
        if (uidx < 0) {  // user id not in DB
            rsp->error = MSG_DB_DEL_FM_FAIL;
            rsp->param2 = MSG_APP_UID_EXIST;
        } else {
            i = *(u16*)(&param[2]); // get FM index of request
            ret = kapp_db_get_fm0_cnt(uidx); // get FM index of DB
            if (ret != i) {
                rsp->error = MSG_DB_DEL_FM_FAIL;
                rsp->param2 = MSG_APP_BAD_INDEX;
                rsp->param2 += ret << 16;  // show current index count
            } else {  // delete FM
                if (kapp_db_delete_FM(uidx, ret) == 0) {
                    rsp->error = 0;
                } else {
                    rsp->error = MSG_DB_DEL_FM_FAIL;  // DB error occurred
                }
                rsp->param2 = 0;
            }
        }
    } else { // unknown DB action
        rsp->error = MSG_DB_BAD_PARAM;
        rsp->param2 = 0;
    }

    kcomm_send_rsp(CMD_SFID_EDIT_DB, NULL, 0);
}

static int handle_fid_cmd(MsgHdr *hdr, CmdPram *args, RspPram *rsp, void *data)
{
    u16 cmd = hdr->cmd;
    u32 arg1 = args->param1;
    u32 arg2 = args->param2;
    void *arg3 = args->data;

    switch (cmd) {
    /* kai2do: why so many cmds here? */
    case CMD_QUERY_APPS:
    case CMD_SELECT_APP:
    case CMD_SET_MODE:
    case CMD_SET_EVENTS:
    case CMD_UPDATE:
    case CMD_ABORT:
    case CMD_SFID_LW3D_START:
    case CMD_SFID_START:
        handle_fid_start(rsp, hdr, *(float*)&arg1, *(float*)&arg2, (u8*)arg3);
        break;
    case CMD_SFID_NEW_USER:
        handle_fid_new_user(rsp, arg1, arg2);
        break;
    case CMD_SFID_SEND_IMAGE:
        handle_send_img(rsp, arg1, arg2);
        break;
    case CMD_SFID_LW3D_IMAGE:
        handle_send_lw3d_img(rsp, arg1, arg2, (u16*)arg3);
        break;
    case CMD_SFID_ADD_DB:
        handle_sfid_add_db(rsp);
        break;
    case CMD_SFID_DELETE_DB:
        handle_sfid_del_db(rsp, arg1);
        break;
    case CMD_SFID_EDIT_DB:
        handle_sfid_edit_db(rsp, arg1, (u8*)arg3);
        break;
    default:
        return -1;
    }

    return 0;
}

static void init_cmd_handler()
{
    init_img_cfg();
}

/* ############################
 * ##    public functions    ##
 * ############################ */

void kcomm_enable_fid_cmds(void)
{
    init_cmd_handler();
    kcomm_reg_cmd_handler(handle_fid_cmd);
}
