/*
 * Kneron Host Communication Image Streaming Inference (ISI) driver
 *
 * Copyright (C) 2020 Kneron, Inc. All rights reserved.
 *
 */
#include <stdio.h>
#include <string.h>
#include "cmsis_os2.h"
#include "base.h"
#include "def_share.h"
#include "com.h"
#include "com_err.h"
#include "msg.h"
#include "kmdw_console.h"
#include "kcomm.h"
#include "kmdw_memory.h"
#include "usb_com.h"
#include "kdrv_clock.h"

#define DEF_LOG_CATEG  "kcomm"

// flags for isi inference thread
#define FLAG_ISI_INF_START0  BIT(0)
#define FLAG_ISI_INF_START1  BIT(1)
#define FLAG_ISI_INF_START2  BIT(2)
#define FLAG_ISI_INF_START_ALL (FLAG_ISI_INF_START0 | FLAG_ISI_INF_START1 | FLAG_ISI_INF_START2)

#define FLAG_ISI_RESULT_EVT    BIT0
#define FLAG_ISI_ABORT_EVT     BIT15

struct img_buf_ctrl_s {
    uint16_t status;   // buffer status
    uint16_t sequence; // image time stamps
    uint32_t id;       // image id
};

volatile struct img_buf_ctrl_s img_buffer[ISI_BUFFER_MAX];
static uint32_t image_pool;
static uint32_t isi_out;  // ddr for relocation of isi data structure
static uint32_t img_time = 0;
static uint32_t isi_out_size = 0;

#define IMG_POOL(n)     (image_pool + (MAX_IMG_SIZE * n))
#define OUT_POOL(n)     (isi_out + (isi_out_size * n))

osMutexId_t image_buffer_mutex;
const osMutexAttr_t ISI_Mutex_attr = {
  "isi_img_buffer",    // human readable mutex name
  osMutexPrioInherit,  // attr_bits
  NULL,                // memory for control block
  0U                   // size for control block
};

osThreadId_t tid_isi_infr;

osEventFlagsId_t isi_result_evt;
static uint32_t result_evt[ISI_BUFFER_MAX];
static uint32_t result_evt_all;

struct kapp_ops         *isi_ops;
struct kapp_img_run_s   img_run[ISI_BUFFER_MAX];

/* ###########################
 * ##    local functions    ##
 * ########################### */

static bool valid_app_id(int id)
{
    return ((id > APP_UNKNOWN) && (id < APP_ID_MAX));
}

/* image buffer control */
static int find_next_buffer(void)
{
    int available = -1;
    uint32_t earliest = 0xff0000;
    int roll_over = 1;
    int i;
    // check for img id conflict ???
    osMutexAcquire(image_buffer_mutex, osWaitForever);
    for (i = 0; i < ISI_BUFFER_MAX; i++) {
        if (img_buffer[i].sequence > 0x7fff)
            roll_over++;
        if (img_buffer[i].status == 16) {
            if (img_buffer[i].sequence < earliest) {
                earliest = img_buffer[i].sequence;
                available = i;
            }
        }
    }
    if (roll_over > ISI_BUFFER_MAX) {  // all image sequences need to roll over
        for (i = 0; i < ISI_BUFFER_MAX; i++)
            img_buffer[i].sequence &= 0x7fff;
        img_time &= 0x7fff;  // reset the time stamp
    }
    osMutexRelease(image_buffer_mutex);
    return available;
}

static int find_available_buffer(void)
{
    int available = 0;
    osMutexAcquire(image_buffer_mutex, osWaitForever);
    for (int i = 0; i < ISI_BUFFER_MAX; i++) {
        if (img_buffer[i].status == 16)
            available++;
    }
    osMutexRelease(image_buffer_mutex);
    return available;
}

static int find_image_id(u32 image_id)
{
    int match = -1, time = -1;
    osMutexAcquire(image_buffer_mutex, osWaitForever);
    for (int i = 0; i < ISI_BUFFER_MAX; i++) {
        if (img_buffer[i].id == image_id) {
            if (img_buffer[i].sequence > time) {  // select the latest copy
                match = i;
                time = img_buffer[i].sequence;
            }
        }
    }
    osMutexRelease(image_buffer_mutex);
    return match;
}

static int get_earliest_image(int ready_image)
{
    int i, idx = -1;
    uint32_t earliest = 0xff0000;
    osMutexAcquire(image_buffer_mutex, osWaitForever);
    for (i = 0; i < ISI_BUFFER_MAX; i++) {
        if (ready_image & 1 << i) {
            if ((img_buffer[i].sequence < earliest) && (img_buffer[i].status < 16)) {
                earliest = img_buffer[i].sequence;
                idx = i;
            }
        }
    }
    osMutexRelease(image_buffer_mutex);
    return idx;
}

// Inference & result handling
static void handle_result(int idx)
{
    osEventFlagsSet(isi_result_evt, result_evt[idx]);
}

static void handle_inference(int idx)
{
    // idx is 0-n
    if (img_run[idx].img_cfg.image_format & IMAGE_FORMAT_PARALLEL_PROC) {
        dbg_msg("ISI inf[%d] id %x (parallel)\n", idx, img_buffer[idx].id);
        img_run[idx].evt_id = isi_result_evt;
        img_run[idx].evt_flag = result_evt[idx];
    } else {
        dbg_msg("ISI inf[%d] id %x (non parallel)\n", idx, img_buffer[idx].id);
        img_run[idx].evt_id = 0;
        img_run[idx].evt_flag = 0;
    }

    (void) isi_ops->run_image(APP_id, &img_run[idx]);

    if (!(img_run[idx].img_cfg.image_format & IMAGE_FORMAT_PARALLEL_PROC)) {
        handle_result(idx);
    }
}

// isi command handler
static void handle_isi_start(RspPram *rsp, u32 app_id, u32 mem_size, u16 img_width, u16 img_height, u32 img_format)
{
    dbg_msg("[kcomm] Start ISI mode.\n");

    if (!valid_app_id(app_id) || (isi_ops == NULL) || (isi_ops->start(app_id) != KAPP_OP_OK)) {
        err_msg("APP_id %d is not supported.\n", app_id);
        rsp->error = MSG_CONFIG_ERROR;
        rsp->param2 = 0;
        kcomm_send_rsp(CMD_ISI_START, NULL, 0);
        return;
    }

    // Allocate buffer memory for results according to the mem size from host
    APP_id = app_id;
    isi_out_size = mem_size;

    if (!isi_out)
        isi_out = kmdw_ddr_reserve(isi_out_size * ISI_BUFFER_MAX);

    if (isi_out) {  // make sure we have results buffer
        rsp->error = NO_ERROR;
        rsp->param2 = ISI_BUFFER_MAX;
        memset((void*)&img_buffer, 0, sizeof(img_buffer));
        for(int i = 0; i < ISI_BUFFER_MAX; i++) {
            img_run[i].img_cfg.image_col = img_width;  // set up rgb image format
            img_run[i].img_cfg.image_row = img_height;
            img_run[i].img_cfg.image_ch = 3;
            img_run[i].img_cfg.image_format = img_format;
            img_run[i].out = (void *)OUT_POOL(i);
            img_run[i].out_size = isi_out_size;
            img_buffer[i].status = 16;  // mark all image as done

            memset((void*)OUT_POOL(i), 0, isi_out_size);

            result_evt[i] = FLAG_ISI_RESULT_EVT << i;
            result_evt_all |= result_evt[i];
        }
        kcomm_send_rsp(CMD_ISI_START, NULL, 0);
    } else {
        rsp->error = MSG_CONFIG_ERROR;
        rsp->param2 = 0;
        kcomm_send_rsp(CMD_ISI_START, NULL, 0);
    }
}

static void handle_isi_send_img(RspPram *rsp, u32 img_size, u32 img_id)
{
    int early_abort = 1;
    int ret, idx;

    if (!valid_app_id(APP_id)) {
        rsp->error = MSG_CONFIG_ERROR;
    } else {
        idx = find_next_buffer();
        if (idx < 0)
            rsp->error = MSG_DME_Q_FULL;  // can't accept another image
        else
            early_abort = 0;
    }

    if (early_abort) {
        rsp->param2 = CMD_ISI_SEND_IMAGE;  // add source id
        kcomm_send_rsp(CMD_ACK_NACK, NULL, 0);
        return;
    }
    u32 remaining = find_available_buffer();
    ret = kcomm_read(IMG_POOL(idx), img_size);
    if (ret != img_size) { // if image not received
        rsp->error = MSG_APP_BAD_IMAGE;
        rsp->param2 = remaining;
        *(u32*)rsp->data = (u32) ret;  // include debug data in reserved field
        kcomm_send_rsp(CMD_ISI_SEND_IMAGE, rsp->data, 4);
        return;
    }
    img_time++;
    img_run[idx].img_cfg.image_mem_addr = IMG_POOL(idx);
    img_run[idx].img_cfg.image_buf_active_index = idx;
    img_buffer[idx].id = img_id;
    img_buffer[idx].status = 1;
    img_buffer[idx].sequence = img_time;

    memset((void*)OUT_POOL(idx), 0, isi_out_size);

    osThreadFlagsSet(tid_isi_infr, 1 << idx);  // start inference

    rsp->error = NO_ERROR;
    rsp->param2 = remaining - 1;
    *(u32*)rsp->data = 0;
    kcomm_send_rsp(CMD_ISI_SEND_IMAGE, rsp->data, 4);
}
extern u32 get_result_wait;
static void handle_isi_get_results(RspPram *rsp, u32 img_id)
{
    int have_result = 0;
    void *out;
    u32 out_size;
    int flags;
    int idx;

    get_result_wait = 0;
    rsp->param2 = img_id;
    *(u32*)rsp->data = 0;
    if (!valid_app_id(APP_id)) {
        rsp->error = MSG_CONFIG_ERROR;
    } else {  // search for the image id
        idx = find_image_id(img_id);
        if (idx < 0) {
            err_msg("[ISI] find_image_id fail: %d: %d\n", idx, img_id);
            rsp->error = MSG_DME_BAD_SSID;
        } else {
            get_result_wait = 1;
            usb_set_abort_flag(isi_result_evt, FLAG_ISI_ABORT_EVT);
            flags = osEventFlagsWait(isi_result_evt, result_evt[idx] | FLAG_ISI_ABORT_EVT,
                             osFlagsNoClear, osWaitForever);
            if (flags & FLAG_ISI_ABORT_EVT) {
                err_msg("[ISI] USB Abort [%08x] (%d[%d])\n", flags, result_evt[idx], idx);
                get_result_wait = 0;
                kdrv_delay_us(500*1000);  // delay 500mS to get all inference completed
                osEventFlagsClear(isi_result_evt, 0xff);  // clear ALL pending events
                for (idx = 0; idx < ISI_BUFFER_MAX; idx++)
                    img_buffer[idx].status = 16;  // reset all image status
                return;
            }
            if (flags != result_evt[idx])
                dbg_msg("[ISI] 1+ events 0x%08x (%d[%d] expected)\n", flags, result_evt[idx], idx);
            osEventFlagsClear(isi_result_evt, result_evt[idx]);

            get_result_wait = 0;
            img_buffer[idx].status = 16;

            // get detection results
            out = (void *)OUT_POOL(idx);
            rsp->error = 0;

            // calculate output size
            out_size = isi_ops->get_result_len(APP_id, out);

            *(u32*)rsp->data = out_size;
            have_result = 1;
        }
    }
    kcomm_send_rsp(CMD_ISI_GET_RESULTS, rsp->data, 4);
    if (have_result)
        kcomm_write((u8*)OUT_POOL(idx), out_size, WMODE_ALT);
}

static int handle_isi_cmd(MsgHdr *hdr, CmdPram *args, RspPram *rsp, void *data)
{
    u16 cmd = hdr->cmd;
    u32 arg1 = args->param1;
    u32 arg2 = args->param2;

    switch (cmd) {
    case CMD_ISI_START:
        handle_isi_start(rsp, arg1, arg2, *(u16*)&args->data, *(u16*)&args->data[2], *(u32*)&args->data[4]);
        break;
    case CMD_ISI_SEND_IMAGE: {
        handle_isi_send_img(rsp, arg1, arg2);
        break;
    }
    case CMD_ISI_GET_RESULTS:
        handle_isi_get_results(rsp, arg1);
        break;
    default:
        return -1;
    }

    return 0;
}

/* ############################
 * ##    inference thread    ##
 * ############################ */

static void isi_inference_thread(void *argument)
{
    int32_t flags, next;

    for (;;) {
        flags = osThreadFlagsWait(FLAG_ISI_INF_START_ALL, osFlagsNoClear, osWaitForever);

        if (flags & FLAG_ISI_INF_START_ALL) {
            next = get_earliest_image(flags & FLAG_ISI_INF_START_ALL);
            if (next < 0) {
                err_msg("INF error, events = %d\n", flags);
                osThreadFlagsClear(flags);  // clear all pending flags
            } else {
                osThreadFlagsClear(1 << next);
                handle_inference(next);
                //dbg_msg("INF %d\n", img_buffer[next].sequence);
            }
        } else {
            osThreadFlagsClear(flags & ~FLAG_ISI_INF_START_ALL);
            err_msg("[isi thread] unknown event: %d\n", flags);
        }
    }
}

/* ############################
 * ##    public functions    ##
 * ############################ */

void kcomm_enable_isi_cmds(struct kapp_ops *ops)
{
    // Allocate buffer memory for images
    image_pool = kmdw_ddr_reserve(MAX_IMG_SIZE * ISI_BUFFER_MAX);

    isi_result_evt = osEventFlagsNew(0);
    isi_ops = ops;
    image_buffer_mutex = osMutexNew(&ISI_Mutex_attr);

    kcomm_reg_cmd_handler(handle_isi_cmd);

    osThreadAttr_t attr;
    memset(&attr, 0, sizeof(attr));
    attr.priority = osPriorityHigh;
    tid_isi_infr = osThreadNew(isi_inference_thread, NULL, &attr);
}
