/*
 * Kneron Host Communication driver
 *
 * Copyright (C) 2018 Kneron, Inc. All rights reserved.
 *
 */
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include "base.h"
#include "def_share.h"
#include "com.h"
#include "com_err.h"
#include "msg.h"
#include "kmdw_console.h"
#include "kcomm.h"
#include "kdev_flash.h"
#include "kdrv_clock.h"

#define DEF_LOG_CATEG  "kcomm"

/* ###########################
 * ##    local functions    ##
 * ########################### */

static int flash_wait_ready(int timeout_ms)
{
    kdev_flash_status_t flash_status;
    int i;

    for (i = 0; i < timeout_ms; i++) {
        flash_status = kdev_flash_get_status();
        if (flash_status.busy == 0) break;
        kdrv_delay_us(1*1000);
    }
    if (i == timeout_ms) i = -1;  // we have timed out
    return i;
}

/* Flash sector merge function operates on a single sector at a time
*  input - starting address of original flash data
*        - starting address of new data to be written
*        - size of new data in bytes
*  output - 0 -> data is identical, no write needed
*         - positive -> data is merged, no need to erase first
*         - negative -> data is merged, needs erase first
*/
static int flash_sector_merge(u8* sector_offset, u8* data_start, u32 data_size)
{
    // we don't do data alignment adjustment yet, assume all inputs are 32-bit aligned
    u16 need_erase = 0, need_update = 0;
    u32 i, *pflash_word, *pnew_word;

    pflash_word = (u32*)sector_offset;
    pnew_word = (u32*)data_start;
    for (i = data_size >> 2; i > 0; i--)
    {
        if (*pflash_word != *pnew_word) need_update = 1;
        if ((*pflash_word ^ *pnew_word) & *pnew_word) {
            need_erase = 1;
        }
        else {
            pflash_word++;
            pnew_word++;
        }
        if (need_erase && need_update) break;  // early abort
    }

    if (need_update) {
        memcpy(sector_offset, data_start, data_size);
        if (need_erase) {
            return -1;
        }
        return 1;
    }
    return 0;  // data are identical
}

/* ########################
 * ##    cmd handlers    ##
 * ######################## */

static void handle_test_echo(RspPram *rsp, u32 len, u8 *pdata)
{
    dlog("len [%d]", len);

    rsp->error = NO_ERROR;
    kcomm_send_rsp(CMD_TEST_ECHO, pdata, len);
}

static void handle_file_write(RspPram *rsp, u32 addr, u32 len)
{
    dlog("addr [0x%x], len [%u]", addr, len);

    u32 read = kcomm_read(addr, len);

    if (read != len) {
        err_msg("read failed with ret [%u], expect [%u]", read, len);
        rsp->error = FILE_ERROR;
    } else {
        rsp->error = NO_ERROR;
    }

    rsp->param2 = read;
    kcomm_send_rsp(CMD_FILE_WRITE, NULL, 0);
}

static void handle_flash_mem_write(RspPram *rsp, u32 addr, u32 len, u8 *data)
{
    dlog("addr [0x%x], len [%u], data [0x%x]", addr, len, data);

    // This code is Winbond specific.  For other vendors, modification might be needed!!!!!!
    u32 offset_addr = addr & 0x0FFF;  // find offset addr into sector
    u8* flash_sector_start = msg_tbuf + 0x40;  // keep away from potential data
    u8* flash_sector_check = msg_rbuf + 0x40;  // reuse the buffer
    u32 i;

    if ((len + offset_addr) > 0x1000) {
        rsp->error = 1; // target addr out of range
        rsp->param2 = 0;
    } else {  // peform FLASH data compare, merge and write
        // load flash sector into tx buf as temp storage
        kdev_flash_readdata(addr & 0xFFFFF000, flash_sector_start, 0x1000);  // read in 4KB
        int ret = flash_wait_ready(100);  // wait for the read operation to complete
        dbg_msg("[kcomm] Read Flash Sector Wait Time -> %d\r\n", ret);
        if (ret < 0) {
            err_msg("[ERROR] host_com: Read Flash Sector failed, ret -> %d\r\n", ret);
        }

        ret = flash_sector_merge(flash_sector_start+offset_addr, data, len);
        info_msg("[kcomm] Flash Merge ret code is %d\r\n", ret);

        if (ret < 0) { // erase the sector first before programming the whole 4KB
            kdev_flash_erase_sector(addr);
            ret = flash_wait_ready(100);  // wait for the erase operation to complete
            if (ret < 0) {
                dbg_msg("[kcomm] Erase Flash Sector Wait Time -> %d\r\n", ret);
            } else {
                for (i = 0; i < 0x1000; i += 0x100) {  // program each 256 byte page separately
                    kdev_flash_programdata((addr & 0xFFFFF000)+i, flash_sector_start+i, 0x100);
                    flash_wait_ready(100);
                }
                ret = -1;  // restore ret value for later stages
            }
        }

        if (ret > 0) {  // program the new data only, one page at a time
            u32 first_page = 0x100 - (offset_addr & 0xFF);  // size of first page
            u32 more_page = len - first_page;
            u32 last_page = more_page & 0xFF;
            more_page = more_page & 0xF00;
            if (first_page) { // do we have partial first page?
                kdev_flash_programdata(addr, flash_sector_start + offset_addr, first_page);
                flash_wait_ready(100);
            }
            for (i = first_page; i < first_page + more_page; i += 0x100) {  // program complete pages
                kdev_flash_programdata(addr + i, flash_sector_start + i, 0x100);
                ret = flash_wait_ready(100);
                dbg_msg("[kcomm] Flash Sector Write Wait Time -> %d\r\n", ret);
            }
            if (last_page) {
                i = first_page + more_page;
                kdev_flash_programdata(addr + i, flash_sector_start + i, last_page);
                flash_wait_ready(100);
            }
        }

        // do post check
        if (ret != 0) {
            kdev_flash_readdata(addr & 0xFFFFF000, flash_sector_check, 0x1000);  // read the new sector
            ret = flash_wait_ready(100);  // wait for the read operation to complete
            if (memcmp(flash_sector_start, flash_sector_check, 0x1000)) {  // if not equal, failed verification
                rsp->error = 2;
            } else {
                rsp->error = NO_ERROR;
            }
        } else {
            rsp->error = NO_ERROR;
        }
        rsp->param2 = len;
    }

    kcomm_send_rsp(CMD_FLASH_MEM_WRITE, NULL, 0);
}

static int handle_test_cmd(MsgHdr *hdr, CmdPram *args, RspPram *rsp, void *data)
{
    u16 cmd = hdr->cmd;
    u32 arg1 = args->param1;
    u32 arg2 = args->param2;
    void *arg3 = args->data;

    switch (cmd) {
    case CMD_TEST_ECHO:
        handle_test_echo(rsp, arg2, (u8*)data);
        break;
    case CMD_FILE_WRITE:
        handle_file_write(rsp, arg1, arg2);
        break;
    case CMD_FLASH_MEM_WRITE:
        handle_flash_mem_write(rsp, arg1, arg2, (u8*)arg3);
        break;
    default:
        return -1;
    }

    return 0;
}

/* ############################
 * ##    public functions    ##
 * ############################ */

void kcomm_enable_test_cmds(void)
{
    kcomm_reg_cmd_handler(handle_test_cmd);
}
