/*
 * Kneron Host Communication driver
 *
 * Copyright (C) 2018 Kneron, Inc. All rights reserved.
 *
 */
#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "cmsis_os2.h"
#include "kmdw_console.h"
#include "kcomm.h"
#include "base.h"
#include "msg.h"
#include "version.h"

#define DEF_LOG_CATEG  "kcomm"

#define HANDLER_SIZE  10

static CmdHandler cmd_handlers[HANDLER_SIZE];

/* ############################
 * ##    static functions    ##
 * ############################ */

static int check_msg(u8 *buf, int size)
{
    MsgHdr *hdr = (MsgHdr*) buf;

    if (hdr->preamble != MSG_HDR_CMD) {
        err_msg("BAD Preamble %d\r\n", size);
        return -1;
    }

    if (hdr->ctrl & PKT_CRC_FLAG) {
        int length = hdr->ctrl & 0x01FFF;
        u32 crc16 = gen_crc16(buf, length);

        if (crc16 != (buf[length+0]+(buf[length+1]<<8))) {
            // send CRC ERROR packet
            err_msg("BAD CRC [%d] crc16=%x : %x\r\n", length, crc16, buf[length+3]+(buf[length+3]<<8));

            // kai2do: pack this as a function?
            u16 rsp_cmd = hdr->cmd | 0x8000;
            hdr = (MsgHdr*) msg_tbuf;
            hdr->cmd = rsp_cmd;
            hdr->msg_len = 8;

            RspPram *rsp = (RspPram*) (msg_tbuf + sizeof(MsgHdr));
            rsp->error = CMD_CRC_ERR;
            rsp->param2 = 0;

            kcomm_write_msg(msg_tbuf, sizeof(MsgHdr)+sizeof(RspPram), 0);  // send crc error packet, without crc

            return -1;
        }
    }

    return 0;
}

static int process_msg(u8 *buf, int size)
{
    if (0 != check_msg(buf, size)) {
        return -1;
    }

    int ret = -1;
    MsgHdr *hdr = (MsgHdr*) buf;
    CmdPram *args = (CmdPram*) (buf + sizeof(MsgHdr)); // used to parse operational commands
    RspPram *rsp = (RspPram*) (buf + sizeof(MsgHdr)); // use the same rcv buffer to format responses
    void *data = (buf + sizeof(MsgHdr) + sizeof(CmdPram)); // points to stuff beyond error and byte count
    u32 cmd = hdr->cmd;

    dbg_msg("[kcomm] receive cmd [0x%x]\n", cmd);

    for (int i = 0; i < HANDLER_SIZE; ++i) {
        CmdHandler handler = cmd_handlers[i];
        if (!handler) {
            break;
        }

        ret = handler(hdr, args, rsp, data);
        if (0 == ret) {
            break;
        }
    }

    if (ret != 0) {
        err_msg("[kcomm] cmd [0x%x] not handled\n", cmd);
        rsp->error = RSP_UKNOWN_CMD;
        rsp->param2 = 0;
        kcomm_send_rsp(cmd, NULL, 0);
    }

    return ret;
}

static void kcomm_thread(void *args)
{
    int32_t flags;

    for (;;) {
        flags = osThreadFlagsWait(FLAG_COMM_RX_DONE | FLAG_COMM_INFR1_DONE | FLAG_COMM_INFR2_DONE, osFlagsWaitAny, osWaitForever);

        if (flags & (FLAG_COMM_INFR1_DONE | FLAG_COMM_INFR2_DONE | FLAG_COMM_RX_DONE)) {
            if (flags & FLAG_COMM_INFR1_DONE) {
                kcomm_dme_complete(1);
            }
            if (flags & FLAG_COMM_INFR2_DONE) {
                kcomm_dme_complete(0);
            }
            if (flags & FLAG_COMM_RX_DONE) {
                process_msg(msg_rbuf, msg_rbuf[2]+4);  // only process the first cmd in the buffer
            }
        }
        else {
            err_msg("[kcomm] unknown event: %d\n", flags);
        }
    }
}

/* ############################
 * ##    public functions    ##
 * ############################ */

void kcomm_reg_cmd_handler(CmdHandler handler)
{
    for (int i = 0; i < HANDLER_SIZE; ++i) {
        if (cmd_handlers[i] == NULL) {
            cmd_handlers[i] = handler;
            return;
        }
    }

    err_msg("[kcomm] too many cmd handlers!\n");
    abort();
}

osThreadId_t tid_host_comm;

void kcomm_init()
{
    osThreadAttr_t attr;
    memset(&attr, 0, sizeof(attr));
    attr.priority = osPriorityAboveNormal;
    tid_host_comm = osThreadNew(kcomm_thread, NULL, &attr);
    kcomm_msg_init(tid_host_comm);
}
