/*
 * Kneron Tiny Yolo Application
 *
 * Copyright (C) 2020 Kneron, Inc. All rights reserved.
 *
 */
#include <string.h>
#include <stdlib.h>
#include "cmsis_os2.h"
#include "kmdw_display.h"
#include "kl520_include.h"
#include "kmdw_memory.h"
#include "kdrv_fb_mgr.h"
#include "kmdw_camera.h"
#include "kapp_tiny_yolo_inf.h"
#include "kmdw_console.h"
#include "model_res.h"
#include "kmdw_model.h"
#include "kdrv_ncpu.h"
#include "kapp_ops.h"
#include "tiny_yolo_ops.h"

#define TINY_YOLO_DEBUG
#define MODEL_COMPILATION_WITH_ADD_NORM

#ifdef TINY_YOLO_DEBUG
#define ty_msg(fmt, ...) info_msg(fmt, ##__VA_ARGS__)
#else
#define ty_msg(fmt, ...)
#endif

#define CAMERA_IDX              CAMERA_RGB_IDX
//#define CAMERA_IDX              CAMERA_NIR_IDX

#define TY_MIN_DRAW_PERIOD_MS       50
#define TY_MIN_INF_DELAY_MS         5
#define TY_MAX_IMAGE_BUFS           4
#define TY_MIN_SCORE                (0.3f)
#define TY_MAX_BOX_NUM              50

#define FLAGS_YOLO_RESULT_EVT       BIT0

#if (CAMERA_IDX == CAMERA_NIR_IDX)
#define IMG_W           NIR_IMG_SOURCE_W
#define IMG_H           NIR_IMG_SOURCE_H
#define IMG_CH          1
#define IMG_SIZE        (IMG_W * IMG_H * IMG_CH)
#define IMG_FORMAT      IMAGE_FORMAT_RAW8
#define DISPLAY_FORMAT  V2K_PIX_FMT_RAW8
#ifdef MODEL_COMPILATION_WITH_ADD_NORM
#define INF_IMG_FORMAT  (IMAGE_FORMAT_SUB128 | NPU_FORMAT_NIR | IMAGE_FORMAT_PARALLEL_PROC)
#else
#define INF_IMG_FORMAT  (IMAGE_FORMAT_RIGHT_SHIFT_ONE_BIT | NPU_FORMAT_NIR | IMAGE_FORMAT_PARALLEL_PROC)
#endif
#else
#define IMG_W           RGB_IMG_SOURCE_W
#define IMG_H           RGB_IMG_SOURCE_H
#define IMG_CH          3
#define IMG_SIZE        (IMG_W * IMG_H * IMG_CH)
#define IMG_FORMAT      IMAGE_FORMAT_RGB565
#define DISPLAY_FORMAT  V2K_PIX_FMT_RGB565
#ifdef MODEL_COMPILATION_WITH_ADD_NORM
#define INF_IMG_FORMAT  (IMAGE_FORMAT_SUB128 | NPU_FORMAT_RGB565 | IMAGE_FORMAT_PARALLEL_PROC)
#else
#define INF_IMG_FORMAT  (IMAGE_FORMAT_RIGHT_SHIFT_ONE_BIT | NPU_FORMAT_RGB565 | IMAGE_FORMAT_PARALLEL_PROC)
#endif
#endif

static int yolo_running;
static int camera_inited;
static int display_inited;
static int do_parallel = 1;

osThreadId_t tid_inference = NULL;
osThreadId_t tid_draw = NULL;

osEventFlagsId_t yolo_result_evt;

struct kapp_camera_settings {
    uint32_t width;             /**< Camera image width*/
    uint32_t height;            /**< Camera image hieght*/
    uint32_t pixelformat;       /**< Camera image pixel format*/
} cam_settings[IMGSRC_NUM];

int g_buf_idx[IMGSRC_NUM] = {-1, -1};

typedef struct yolo_result_s_l {
    uint32_t class_count;            // total class count
    uint32_t box_count;              // boxes of all classes
    struct bounding_box_s boxes[TY_MAX_BOX_NUM];  // box_count
} yolo_result_s_l_t;

struct kapp_img_run_s   img_run[TY_MAX_IMAGE_BUFS];

static uint32_t result_evt[TY_MAX_IMAGE_BUFS];
static uint32_t result_evt_all;
static int inf_buf_idx;

static uint32_t start_ticks, end_ticks, inf_count;
static uint32_t seq_num_1st, seq_num;
static float params[MAX_PARAMS_LEN];

struct kapp_ops *ops;

enum {
    person,
    bicycle,
    car,
    motorbike,
    aeroplane,
    bus,
    train,
    truck,
    boat,
    traffic_light,
    fire_hydrant,
    stop_sign,
    parking_meter,
    bench,
    bird,
    cat,
    dog,
    horse,
    sheep,
    cow,
    elephant,
    bear,
    zebra,
    giraffe,
    backpack,
    umbrella,
    handbag,
    tie,
    suitcase,
    frisbee,
    skis,
    snowboard,
    sports_ball,
    kite,
    baseball_bat,
    baseball_glove,
    skateboard,
    surfboard,
    tennis_racket,
    bottle,
    wine_glass,
    cup,
    fork,
    knife,
    spoon,
    bowl,
    banana,
    apple,
    sandwich,
    orange,
    broccoli,
    carrot,
    hot_dog,
    pizza,
    donut,
    cake,
    chair,
    sofa,
    pottedplant,
    bed,
    diningtable,
    toilet,
    tvmonitor,
    laptop,
    mouse,
    remote,
    keyboard,
    cell_phone,
    microwave,
    oven,
    toaster,
    sink,
    refrigerator,
    book,
    clock,
    vase,
    scissors,
    teddy_bear,
    hair_drier,
    toothbrush,
};

/*-----------------------------------------------------------------------------------------
 *                           Private API
 *-----------------------------------------------------------------------------------------*/
static void yolo_check_fps(void)
{
#ifdef TINY_YOLO_DEBUG
    static uint32_t last_time = 0;
    static float fps = 0;
    uint32_t cur_time, elapsed_time;

    if (last_time == 0) {
        // skip first one
        last_time = osKernelGetTickCount();
        return;
    }

    cur_time = osKernelGetTickCount();
    elapsed_time = cur_time - last_time;

    fps = 1000.0f / elapsed_time;
    ty_msg("--> FPS: %3.2f (%d ms)\n", fps, elapsed_time);
    last_time = osKernelGetTickCount();
#endif
}

static void kapp_yolo_draw_result(int buf_idx, struct yolo_result_s_l* yolo_result_p)
{
    u32 buf_addr;
    int i, x, y, w, h, read_idx;
    struct img_run_time_s run_time;

    buf_addr = kdrv_fb_mgr_next_read(CAMERA_IDX, &read_idx);
    kmdw_display_update_draw_fb(buf_addr, CAMERA_IDX);

    ty_msg("[R%8.8d] boxes %d (show > %f) (buf_idx %d)\n", osKernelGetTickCount(),
        yolo_result_p->box_count, TY_MIN_SCORE, buf_idx);

    for (i = 0; i < yolo_result_p->box_count; i++) {
        if (yolo_result_p->boxes[i].score <= TY_MIN_SCORE)
            continue;

        x = yolo_result_p->boxes[i].x1;
        y = yolo_result_p->boxes[i].y1;
        w = yolo_result_p->boxes[i].x2 - x;
        h = yolo_result_p->boxes[i].y2 - y;
#if (CAMERA_IDX == CAMERA_RGB_IDX)
        kmdw_display_set_pen_rgb565(YELLOW, 4);
#else
        kmdw_display_set_pen_rgb565(BLACK, 1);
#endif
        kmdw_display_draw_rect(x, y, w, h, 2);

        ty_msg("box[%d]: class=%d: (%d, %d) (%d, %d) @ score=%f\n", i,
            yolo_result_p->boxes[i].class_num,
            (int)yolo_result_p->boxes[i].x1,
            (int)yolo_result_p->boxes[i].y1,
            (int)yolo_result_p->boxes[i].x2,
            (int)yolo_result_p->boxes[i].y2,
            yolo_result_p->boxes[i].score);
    }

    kdrv_fb_mgr_read_done(CAMERA_IDX, read_idx);

    kmdw_model_get_run_time(buf_idx, &run_time);

    ty_msg("Round trip %d: pre/npu/post: %d/%d/%d ms\n",
        run_time.round_trip_time, run_time.pre_proc_time,
        run_time.npu_proc_time, run_time.post_proc_time);
}

static void kapp_camera_setting(uint8_t cam_idx, unsigned int width, unsigned int height, unsigned int pixelformat)
{
    cam_settings[cam_idx].width = width;
    cam_settings[cam_idx].height = height;
    cam_settings[cam_idx].pixelformat = pixelformat;
}

static void kapp_camera_open(uint8_t cam_idx)
{
    struct cam_format fmt;

    memset(&fmt, 0, sizeof(fmt));

    (void) kmdw_camera_open(cam_idx);

    fmt.width = cam_settings[cam_idx].width;
    fmt.height = cam_settings[cam_idx].height;
    fmt.pixelformat = cam_settings[cam_idx].pixelformat;
    (void) kmdw_camera_set_frame_format(cam_idx, (struct cam_format *)&fmt);

    (void) kmdw_camera_get_frame_format(cam_idx, &fmt);

    (void) kmdw_camera_buffer_init(cam_idx);
}

static void kapp_camera_start(unsigned int cam_idx)
{
    (void) kmdw_camera_start(cam_idx);
}

static void kapp_camera_stop(unsigned int cam_idx)
{
    (void) kmdw_camera_stop(cam_idx);
}

static void kapp_camera_close(unsigned int cam_idx)
{
    (void) kmdw_camera_close(cam_idx);
}

static void kapp_camera_init(void)
{
    if (camera_inited == 0) {
        kapp_camera_open(CAMERA_IDX);
        kapp_camera_start(CAMERA_IDX);

        camera_inited = 1;
    }
}

static void kapp_camera_exit(void)
{
    if (camera_inited == 1) {
        kapp_camera_stop(CAMERA_IDX);
        kapp_camera_close(CAMERA_IDX);

        camera_inited = 0;
    }
}

static void kapp_display_init(void)
{
    struct video_input_params params;

    if (display_inited == 0) {
        params.input_fmt = DISPLAY_FORMAT;
        params.input_xres = IMG_W;
        params.input_yres = IMG_H;
        kmdw_video_renderer_open(&params);
        kmdw_video_renderer_set_camera(CAMERA_IDX);
        kmdw_display_set_pen_rgb565(BLACK, 1);
        kmdw_video_renderer_buffer_initialize(&params);
        kmdw_video_renderer_start();
        display_inited = 1;
    }
}

static void kapp_display_exit(void)
{
    if (display_inited == 1) {
        kmdw_video_renderer_stop();
        display_inited = 0;
    }
}

static uint32_t yolo_fb_capture(void)
{
    int cam_idx;
    uint32_t frame_addr;

#ifdef TINY_YOLO_DEBUG
    uint32_t pre = osKernelGetTickCount();
#endif

    cam_idx = CAMERA_IDX;
    frame_addr = kdrv_fb_mgr_next_inf(cam_idx, &g_buf_idx[cam_idx]);
    seq_num = kdrv_fb_mgr_get_buf_seq_num(cam_idx, g_buf_idx[cam_idx]);
    if (seq_num_1st == (uint32_t)-1)
        seq_num_1st = seq_num;

    ty_msg("\n=[%8.8d]= [%d] camera # %d (pre %8.8d)\n", osKernelGetTickCount(), inf_buf_idx, seq_num, pre);

    return frame_addr;
}

static void yolo_fb_prepare(void)
{
    int cam_idx;

    cam_idx = CAMERA_IDX;
    kdrv_fb_mgr_inf_done(cam_idx, g_buf_idx[cam_idx]);
    ty_msg("-[%8.8d]- (inf out)\n", osKernelGetTickCount());
}

static void yolo_fb_free(void)
{
    kdrv_fb_mgr_free_inf_buf(CAMERA_IDX);
}

void model_app_tiny_yolo(uint32_t img_mem_addr)
{
    int i = inf_buf_idx;

    img_run[i].img_cfg.image_mem_addr = img_mem_addr;
    img_run[i].img_cfg.image_buf_active_index = i;

    if (do_parallel) {
        img_run[i].img_cfg.image_format |= IMAGE_FORMAT_PARALLEL_PROC;
        img_run[i].evt_id = yolo_result_evt;
        img_run[i].evt_flag = result_evt[i];
    } else {
        img_run[i].img_cfg.image_format &= ~IMAGE_FORMAT_PARALLEL_PROC;
        img_run[i].evt_id = 0;
        img_run[i].evt_flag = 0;
    }

    ops->run_image(APP_OD, &img_run[i]);

    if (!do_parallel) {
        kapp_yolo_draw_result(i, img_run[i].out);
        yolo_check_fps();
    }

    // Prepare for next
    inf_buf_idx = (inf_buf_idx + 1) % TY_MAX_IMAGE_BUFS;
}

static void kapp_yolo_model_run(void)
{
    uint32_t addr;

    if (yolo_running) {
        addr = yolo_fb_capture();

        model_app_tiny_yolo(addr);

        yolo_fb_prepare();
    }
}

static bool kapp_yolo_model_init(void)
{
    yolo_result_evt = osEventFlagsNew(0);

    params[0] = TY_MIN_SCORE;

    for (int i = 0; i < TY_MAX_IMAGE_BUFS; i++) {
        img_run[i].out = (struct yolo_result_s_l*)kmdw_ddr_reserve(sizeof(struct yolo_result_s_l));

        result_evt[i] = FLAGS_YOLO_RESULT_EVT << i;
        result_evt_all |= result_evt[i];

        img_run[i].img_cfg.image_mem_len = IMG_SIZE;
        img_run[i].img_cfg.image_col = IMG_W;
        img_run[i].img_cfg.image_row = IMG_H;
        img_run[i].img_cfg.image_ch = IMG_CH;
        img_run[i].img_cfg.image_format = INF_IMG_FORMAT;

        img_run[i].ext_param = params;
    }

    kapp_camera_setting(CAMERA_IDX, IMG_W, IMG_H, IMG_FORMAT);

    return TRUE;
}

static void yolo_inference_start(void)
{
    yolo_running = 1;
}

static void yolo_inference_stop(void)
{
    yolo_fb_free();
    yolo_running = 0;
}

static void kapp_yolo_draw_thread(void *argument)
{
    uint32_t flags, evt;

    while(1)
    {
        flags = osEventFlagsWait(yolo_result_evt, result_evt_all,
                             osFlagsWaitAny, osWaitForever);
   
        evt = FLAGS_YOLO_RESULT_EVT;
        for (int i = 0; i < TY_MAX_IMAGE_BUFS; i++) {
            if (flags && (flags & (evt << i))) {
                kapp_yolo_draw_result(i, img_run[i].out);
                yolo_check_fps();
                flags &= ~(evt << i);
                continue;
            }
        }
    }
}

static void kapp_tiny_yolo_thread(void *argument)
{
    uint32_t flags, timeout = osWaitForever;

    kapp_yolo_model_init();

    for (;;) {
        flags = osThreadFlagsWait(FLAGS_YOLO_ALL_EVENTS, osFlagsWaitAny, timeout);

        if (flags == osFlagsErrorTimeout) {
            kapp_yolo_model_run();
            inf_count++;
            continue;
        }
        
        if (flags & FLAGS_YOLO_STOP_EVT) {
            kapp_display_exit();
            kapp_camera_exit();
            flags &= ~(FLAGS_YOLO_STOP_EVT | FLAGS_YOLO_START_EVT);

            yolo_inference_stop();
            timeout = osWaitForever;
            end_ticks = osKernelGetTickCount();
            if (inf_count) {
                uint32_t elapsed_ticks, seq_num_diff;
                float avg_elapsed_time, avg_fps;

                if (end_ticks >= start_ticks)
                    elapsed_ticks = end_ticks - start_ticks;
                else
                    elapsed_ticks = 0xffffffff - (start_ticks - end_ticks -1);
                avg_elapsed_time = (float)elapsed_ticks / inf_count;
                avg_fps = 1000.0f / avg_elapsed_time;

                seq_num_diff = seq_num - seq_num_1st + 1;

                info_msg("\n=> Avg %3.2f FPS (%3.2fms = %d/%d, cam frames %d)\n",
                    avg_fps, avg_elapsed_time, elapsed_ticks, inf_count, seq_num_diff);
            }
        }

        if (flags & FLAGS_YOLO_START_EVT) {
            kapp_camera_init();
            kapp_display_init();
            flags &= ~FLAGS_YOLO_START_EVT;

            yolo_inference_start();
            timeout = TY_MIN_INF_DELAY_MS;
            start_ticks = osKernelGetTickCount();
            inf_count = 0;
            seq_num_1st = (uint32_t)-1;
        }
            
        if (flags & FLAGS_YOLO_TOGGLE_EVT) {
            do_parallel = !do_parallel;
            info_msg("Yolo thread: do_parallel %d\n", do_parallel);
            flags &= ~FLAGS_YOLO_TOGGLE_EVT;
        }

        if (flags) {
            err_msg("Yolo thread: unknown flags 0x%x\n", flags);
        }
    }
}

/*----------------------------------------------------------------------------------------
 *                                Public API
 *----------------------------------------------------------------------------------------*/

osThreadId_t kapp_tiny_yolo_inf_init(void)
{
    if (tid_inference != NULL)
        return tid_inference;

    ops = tiny_yolo_get_ops();

    tid_inference = osThreadNew(kapp_tiny_yolo_thread, NULL, NULL);

    tid_draw = osThreadNew(kapp_yolo_draw_thread, NULL, NULL);
    osThreadSetPriority(tid_draw, osPriorityHigh);

    kdrv_ncpu_set_ncpu_debug_lvl(LOG_PROFILE);

    return tid_inference;
}

