#include "board_kl520.h"
#include "console_example.h"
#include <string.h>
#include <stdlib.h>
#include <cmsis_os2.h>
#include "utility.h"
#include "kmdw_camera.h"
#include "kmdw_display.h"
#include "kmdw_console.h"
#include "kmdw_memxfer.h"
#include "kmdw_memory.h"
#include "framework/event.h"
#include "framework/v2k.h"
#include "framework/framework_driver.h"
#include "kdrv_uart.h"
#include "kdrv_i2c.h"
#include "kdrv_system.h"
#include "kdrv_clock.h"
#include "kdp_ddr_table.h"
#include "kdrv_fb_mgr.h"
#include "io.h"
#include "power_manager.h"

#define FLAGS_SYSTEM_TEST_START_EVT                 BIT0
#define FLAGS_SYSTEM_TEST_STOP_EVT                  BIT1
#define FLAGS_SYSTEM_TEST_ALL_EVT                   (FLAGS_SYSTEM_TEST_START_EVT | FLAGS_SYSTEM_TEST_STOP_EVT)

typedef void (*console_cmd_func)(void);
struct console_cmd_info
{
    char *desc;
    console_cmd_func func;
};

struct camera_settings {
    uint32_t width;             /**< Camera image width*/
    uint32_t height;            /**< Camera image hieght*/
    uint32_t pixelformat;       /**< Camera image pixel format*/
} cam_settings[IMGSRC_NUM];

osThreadId_t tid_console_example = 0;
osThreadId_t tid_system_test = 0;

#define CAMERA_RGB_IDX              (MIPI_CAM_RGB)
#define CAMERA_NIR_IDX              (MIPI_CAM_NIR)
#define CAMERA_RGB_IMG_WIDTH        (640)
#define CAMERA_RGB_IMG_HEIGHT       (480)
#define CAMERA_NIR_IMG_WIDTH        (480)
#define CAMERA_NIR_IMG_HEIGHT       (640)
#define LCD_WIDTH                   LCDC_WIDTH
#define LCD_HEIGHT                  LCDC_HEIGHT

static int8_t display_cam = CAMERA_RGB_IDX;
static bool test_en = FALSE;

struct console_cmd_info app_sample_array[] = {
    {"Start system On/Off test",        sample_system_on_off_test_start  },
    {"Stop system On/Off test",         sample_system_on_off_test_stop  },
    {"Quit",                            sample_app_quit  },
};

void sample_init_mipi_camera(void)
{
    if(kdrv_i2c_initialize(KDRV_I2C_CTRL_0, KDRV_I2C_SPEED_400K) == KDRV_STATUS_OK) {
        if(kmdw_camera_init() != KDRV_STATUS_OK) {
            err_msg("Fails to initialize camera\n");
        }
    }
    else {
        err_msg("Fails to initialize I2C driver\n");
    }
}

void sample_camera_setting(uint8_t cam_idx, uint32_t width, uint32_t height, uint32_t pixelformat)
{
    cam_settings[cam_idx].width = width;
    cam_settings[cam_idx].height = height;
    cam_settings[cam_idx].pixelformat = pixelformat;
}

void sample_camera_open(uint8_t cam_idx)
{
    struct cam_format fmt;

    if(cam_idx == MIPI_CAM_RGB) {
        sample_camera_setting(cam_idx, IMGSRC_0_WIDTH, IMGSRC_0_HEIGHT, IMGSRC_0_FORMAT);
    }
    else if(cam_idx == MIPI_CAM_NIR) {
        sample_camera_setting(cam_idx, IMGSRC_1_WIDTH, IMGSRC_1_HEIGHT, IMGSRC_1_FORMAT);
    }
    else {
        return;
    }

    memset(&fmt, 0, sizeof(fmt));
    if(kmdw_camera_open(cam_idx) != 0) {
        err_msg("Fails to open camera\n");
        return;
    }        
    fmt.width = cam_settings[cam_idx].width;
    fmt.height = cam_settings[cam_idx].height;
    fmt.pixelformat = cam_settings[cam_idx].pixelformat;
    (void) kmdw_camera_set_frame_format(cam_idx, (struct cam_format *)&fmt);
    (void) kmdw_camera_get_frame_format(cam_idx, &fmt);
    (void) kmdw_camera_buffer_init(cam_idx);
    (void) kmdw_camera_start(cam_idx);
}

void sample_camera_close(uint8_t cam_idx)
{    
    if((cam_idx != MIPI_CAM_RGB) || (cam_idx != MIPI_CAM_NIR)) {
        return;
    }

    if(kmdw_camera_stop(cam_idx) != KMDW_STATUS_OK) {
        err_msg("Fails to stop camera\n");
        return;
    }
    
    if (kmdw_camera_close(cam_idx) != KMDW_STATUS_OK) {
        err_msg("Fails to close camera\n");
        return;
    }
    kdrv_fb_mgr_free_write_buf(cam_idx);
    kdrv_fb_mgr_free_inf_buf(cam_idx);
}

void sample_open_video_renderer(uint8_t cam_idx, uint32_t width, uint32_t height, uint32_t pixelformat)
{
    struct video_input_params params;

    if(kmdw_display_initialize() == KMDW_STATUS_OK) {
        params.input_fmt = pixelformat;
        params.input_xres = width;
        params.input_yres = height;

        (void)kmdw_video_renderer_open(&params);
        (void)kmdw_video_renderer_set_camera(cam_idx);
        (void)kmdw_video_renderer_buffer_initialize(&params);
        (void)kmdw_video_renderer_start();
    }
    else {
        err_msg("Fails to initialize display\n");
    }
}

void sample_close_video_renderer(void)
{
    kmdw_video_renderer_stop();
}

void sample_switch_video_renderer_src(void)
{
    display_cam = !display_cam;
    DSG("display_cam:%d", display_cam);
    if(CAMERA_RGB_IDX == display_cam) {
        sample_close_video_renderer();
        sample_open_video_renderer(display_cam, CAMERA_RGB_IMG_WIDTH, CAMERA_RGB_IMG_HEIGHT, V2K_PIX_FMT_RGB565);
    }
    else if(CAMERA_NIR_IDX == display_cam) {
        sample_close_video_renderer();
        sample_open_video_renderer(display_cam, CAMERA_NIR_IMG_WIDTH, CAMERA_NIR_IMG_HEIGHT, V2K_PIX_FMT_RAW8);
    }
}

void sample_app_quit(void)
{
    power_manager_shutdown();
}

void sample_system_on_off_test(void)
{
    while(test_en == TRUE)
    {
        DSG(">>>>>>>>>>>> Open system...");    
        sample_camera_open(CAMERA_RGB_IDX);
        sample_camera_open(CAMERA_NIR_IDX);

        display_cam = CAMERA_RGB_IDX;
        sample_open_video_renderer(display_cam, LCD_WIDTH, LCD_HEIGHT, V2K_PIX_FMT_RGB565); 
        
        kdrv_delay_us(500000);
        sample_switch_video_renderer_src();
        kdrv_delay_us(500000);
        
        DSG(">>>>>>>>>>>> Close system...");
        sample_camera_close(CAMERA_RGB_IDX);
        sample_camera_close(CAMERA_NIR_IDX);

        sample_close_video_renderer();
        
        if(test_en == FALSE)
            break;
        kdrv_delay_us(500000);
    }
}

void sample_system_on_off_test_start(void)
{
    test_en = TRUE;
    set_thread_event(tid_system_test, FLAGS_SYSTEM_TEST_START_EVT);
}

void sample_system_on_off_test_stop(void)
{
    test_en = FALSE;
    set_thread_event(tid_system_test, FLAGS_SYSTEM_TEST_STOP_EVT);
}


void console_thread(void *arg)
{
    int id;
    int i = 0;
    uint32_t rc = 1;
    uint32_t idle_cmd_size = ARRAY_SIZE(app_sample_array);
    char buf[64];
    console_cmd_func cmd_func;
    
    while(1)
    {
        id = 0;
        if (rc)
            goto cmd_prompt;
        
        DSG("\n === System On/Off test === ");
        for (i = 0; i < idle_cmd_size; ++i)
        { 
            sprintf(buf, "(%2d) %s", i + 1, app_sample_array[i].desc);
            DSG("%s", buf);
        }

cmd_prompt:
        DSG_NOLF(" command >> ");
        rc = kmdw_console_echo_gets(buf, sizeof(buf));
        if (!rc)
            continue;

        id = atoi(strtok(buf, " \r\n\t"));
        if (id > 0 && id <= idle_cmd_size)
        {
            cmd_func = app_sample_array[id - 1].func;
            if (cmd_func) {
                cmd_func();    
            }
        }
        else {
            err_msg("Invalid command: %d", id);
            continue;
        }
    }
}

static void kdp_system_test_thread(void *argument)
{
     uint32_t flags;

     sample_init_mipi_camera();
     
     for (;;) {
        flags = osThreadFlagsWait(FLAGS_SYSTEM_TEST_ALL_EVT, osFlagsWaitAny, osWaitForever);
        osThreadFlagsClear(flags);

        if (flags == FLAGS_SYSTEM_TEST_START_EVT) {
            DSG("Start system on/off test");
            sample_system_on_off_test();
            flags &= ~(FLAGS_SYSTEM_TEST_START_EVT);
            continue;
        }
        else if (flags == FLAGS_SYSTEM_TEST_STOP_EVT) {
            DSG("Stop system on/off test");
            flags &= ~(FLAGS_SYSTEM_TEST_STOP_EVT);
            continue;
        }
        else {
            err_msg("%s: unknown flags 0x%x\n", flags);
        }
     }
}

int kdp_system_test_thread_create(void)
{
    int ret = -1;
    if (tid_system_test != NULL)
        return ret;

    if((tid_system_test = osThreadNew(kdp_system_test_thread, NULL, NULL)) != NULL) {
        osThreadSetPriority(tid_system_test, osPriorityNormal);
        ret = 0;
    }
    return ret;
}

void console_example_entry(void)
{
    if(kdp_system_test_thread_create() != 0) {
        err_msg("[ERROR] Fail to create kdp_system_test_thread");
        return;
    }
    tid_console_example = osThreadNew(console_thread, NULL, NULL);
    osThreadSetPriority (tid_console_example, osPriorityBelowNormal);
}

